/*
 *  This file is part of the X10 project (http://x10-lang.org).
 *
 *  This file is licensed to You under the Eclipse Public License (EPL);
 *  You may not use this file except in compliance with the License.
 *  You may obtain a copy of the License at
 *      http://www.opensource.org/licenses/eclipse-1.0.php
 *
 *  (C) Copyright IBM Corporation 2006-2014.
 */

package de.uks.plm.ftglb.actor;

import x10.xrx.Runtime;

/**
 * <p>Class that collects lifeline statistics of GLB
 * </p>
 */
public final class Logger
{
    /* workload sent/recieved stat*/
    public var nodesCount:Long = 0;
    public var nodesGiven:Long = 0;
    public var lifelineNodesReceived:Long = 0;

    /* (random)stealing requests stat*/
    public var stealsAttempted:Long = 0;
    public var stealsPerpetrated:Long = 0;
    public var stealsReceived:Long = 0;
    public var stealsSuffered:Long = 0;
    public var nodesReceived:Long = 0;

    /* (lifeline)stealing requests stat*/
    public var lifelineStealsAttempted:Long = 0;
    public var lifelineStealsPerpetrated:Long = 0;
    public var lifelineStealsReceived:Long = 0;
    public var lifelineStealsSuffered:Long = 0;

    /* timing stat */
    public var lastStartStopLiveTimeStamp:Long = -1;
    public var timeAlive:Long = 0;
    public var timeDead:Long = 0;
    public var startTime:Long = 0;
    public val timeReference:Long;

    /* backup stats */
    public var regularBackupsWritten:Long = 0;
    public var stealBackupsWritten:Long = 0;

    /**
     * Constructor
     * @param b true, called when prior-calculation; false, called when
     * post-calculation
     */
    public def this(val b:Boolean) : Logger
    {
        if (b)
        {
            x10.util.Team.WORLD.barrier();
        }

        timeReference = System.nanoTime();
    }

    /**
     * Timer is started before processing, which includes calculation,
     * distribution and requesting/rejects tasks
     */
    public def startLive() : void
    {
        val time = System.nanoTime();
        if (0 == startTime)
        {
            this.startTime = time;
        }
        if (0 <= this.lastStartStopLiveTimeStamp)
        {
            this.timeDead += time - this.lastStartStopLiveTimeStamp;
        }
        this.lastStartStopLiveTimeStamp = time;
    }

    /**
     * Timer is stopped when running out of tasks and failing to steal any task
     */
    public def stopLive() : void
    {
        val time = System.nanoTime();
        this.timeAlive += time - this.lastStartStopLiveTimeStamp;
        this.lastStartStopLiveTimeStamp = time;
    }

    /**
     * Aggregate stats for all places
     * @param logs log from every place
     */
    public def collect(val logs:Rail[Logger]) : void
    {
        for (val l:Logger in logs)
        {
            add(l);
        }
    }

    /**
     * Print out the actual workload re-distribution by showing the steals that
     * were carried out.
     */
    public def stats() :void
    {
        Console.OUT.println(  nodesGiven + " Task items stolen = "
                            + nodesReceived + " (direct) + "
                            + lifelineNodesReceived + " (lifeline).");

        Console.OUT.println(stealsPerpetrated + " successful direct steals.");

        Console.OUT.println(  lifelineStealsPerpetrated
                            + " successful lifeline steals.");

        Console.OUT.println(  regularBackupsWritten + " regular backups written");
        Console.OUT.println(  stealBackupsWritten + " steal backups written");
    }

    /**
     * Gets part of the string.
     * @param str original string
     * @param start starting index of the string
     * @param end ending index of the string
     * @return string from start to end
     */
    static def sub(val str:String, val start:Int, val end:Int) : String
    {
        return (str.substring(start, Math.min(end, str.length())));
    }

    /**
     * Sum up stat with another logger
     * @param other another logger to sum up with
     */
    public def add(other:Logger) : void
    {
        if (null != other)
        {
            this.nodesCount += other.nodesCount;
            this.nodesGiven += other.nodesGiven;
            this.nodesReceived += other.nodesReceived;
            this.stealsPerpetrated += other.stealsPerpetrated;
            this.lifelineNodesReceived += other.lifelineNodesReceived;
            this.lifelineStealsPerpetrated += other.lifelineStealsPerpetrated;
            this.regularBackupsWritten += other.regularBackupsWritten;
            this.stealBackupsWritten += other.stealBackupsWritten;
        }
    }

    /**
     * Print out more detailed lifeline stats when verbose flag turned on
     * @param verbose verbose flag true when {@link GLBParameters} show glb flag
     * is on.
     */
    public def get(val verbose:Boolean) : Logger
    {
        if (verbose)
        {
            Console.OUT.println(  "" + Runtime.hereLong() + " -> "
                                + Logger.sub("" + (timeAlive / 1E9), 0n, 6n)
                                + " : "
                                + Logger.sub("" + (timeDead / 1E9), 0n, 6n)
                                + " : " 
                                + Logger.sub("" + (  (  this.timeAlive
                                                      + this.timeDead)
                                                   / 1E9)
                                                    , 0n
                                                    , 6n)
                                + " : "
                                + Logger.sub(  "" + (  (100.0 * this.timeAlive)
                                                     / (  this.timeAlive
                                                        + this.timeDead))
                                             , 0n
                                             , 6n)
                                + "%" + " :: "
                                + Logger.sub("" + (  (  this.startTime
                                                      - this.timeReference)
                                                   / 1E9)
                                             , 0n
                                             , 6n)
                                + " : "
                                + Logger.sub
                                    ("" + (  (  this.lastStartStopLiveTimeStamp
                                              - this.timeReference)
                                           / 1E9)
                                     , 0n
                                     , 6n)
                                + " :: " + this.nodesCount + " :: "
                                + this.nodesGiven + " : "
                                + this.nodesReceived + " : "
                                + this.lifelineNodesReceived + " :: "
                                + this.stealsReceived + " : "
                                + this.lifelineStealsReceived + " :: "
                                + this.stealsSuffered + " : "
                                + this.lifelineStealsSuffered + " :: "
                                + this.stealsAttempted + " : "
                                + (  this.stealsAttempted
                                   - this.stealsPerpetrated)
                                + " :: " + this.lifelineStealsAttempted + " : "
                                + (  this.lifelineStealsAttempted
                                   - this.lifelineStealsPerpetrated)
                                + " :: " + this.regularBackupsWritten
                                + " : " + this.stealBackupsWritten);
        }
        return (this);
    }
}
