/*
 *  This file is part of the X10 project (http://x10-lang.org).
 *
 *  This file is licensed to You under the Eclipse Public License (EPL);
 *  You may not use this file except in compliance with the License.
 *  You may obtain a copy of the License at
 *      http://www.opensource.org/licenses/eclipse-1.0.php
 *
 *  (C) Copyright IBM Corporation 2006-2014.
 */

package de.uks.plm.ftglb.actor;

import x10.compiler.Inline;
import x10.compiler.Uncounted;

import x10.util.ArrayList;
import x10.util.HashMap;
import x10.util.List;
import x10.util.Option;
import x10.util.OptionsParser;
import x10.util.Pair;
import x10.util.Random;

import x10.xrx.Runtime;

import de.uks.plm.ftglb.actor.dataStructures.BackupData;
import de.uks.plm.ftglb.actor.dataStructures.Restore;
import de.uks.plm.ftglb.actor.dataStructures.Timeout;

import de.uks.plm.ftglb.actor.managers.BackupReceiverManager;
import de.uks.plm.ftglb.actor.managers.BackupSenderManager;
import de.uks.plm.ftglb.actor.managers.LinkManager;
import de.uks.plm.ftglb.actor.managers.MonitorManager;
import de.uks.plm.ftglb.actor.managers.MessageManager;
import de.uks.plm.ftglb.actor.managers.OpenManager;
import de.uks.plm.ftglb.actor.managers.OpenFEndManager;
import de.uks.plm.ftglb.actor.managers.StateManager;
import de.uks.plm.ftglb.actor.managers.TimeoutManager;
import de.uks.plm.ftglb.actor.managers.BackForthManager;
import de.uks.plm.ftglb.actor.managers.ProcessingStackManager;
import de.uks.plm.ftglb.actor.managers.WaitingAndWorkingManager;
import de.uks.plm.ftglb.actor.managers.LogManager;
import de.uks.plm.ftglb.actor.logEvents.*;

import de.uks.plm.utils.AssertPrinter;
import de.uks.plm.utils.SimpleStack;

/**
 * The local runner for the GLB framework. An instance of this class runs at
 * each place and provides the context within which user-specified tasks execute
 * and are load balanced across all places.
 * @param <Queue> Concrete TaskQueue type
 * @param <R> Result type.
 */
public final class Worker[Queue, R] {Queue <: TaskQueue[Queue, R]}
{
    public static val TIMEOUT_REGULAR:Long = 100;
    public static val TIMEOUT_STEAL:Long = 100;
    public static val TIMEOUT_TO:Long = 100;
    public static val TIMEOUT_RST:Long = 100;
    public static val TIMEOUT_LNK:Long = 100;
    public static val TIMEOUT_GOT:Long = 100;
    public static val TIMEOUT_GIVE = 100;
    public static val TIMEOUT_BFACK = 100;
    public static val TIMEOUT_PROCESS_JOINS = 1000;

    /** TaskQueue, responsible for crunching numbers */
    public val queue:Queue;

    /** Lifeline buddies */
    public var lifelines:ArrayList[Long];

    /** The data structure to keep a key invariant: 
     * At any time, at most one message has been sent on an
     * outgoing lifeline (and hence at most one message has
     * been received on an incoming lifeline).*/
    // val lifelinesActivated:Rail[Boolean];

    /** The granularity of tasks to run in a batch before starts to probe
     * network to respond to work-stealing requests. The smaller n is, the more
     * responsive to the work-stealing requests; on the other hand, less focused
     * on computation */
    val n:Int;

    /**
     * Number of random victims to probe before sending requests to lifeline
     * buddy
     */
    val w:Int;

    /** Maximum number of random victims */
    val m:Int;

    /** k * n Cycle to write backups */
    val k:Int;
    var currentK:Int;

    val r:Int;
    var currentR:Int;

    val l:Int;
    var z:Int;

    /** Random number, used when picking a non-lifeline victim/buddy. Important
     * to seed with place id, otherwise BG/Q, the random sequence will be 
     * exactly same at different places
     */
    val random = new Random(Runtime.hereInt());

    /**
     * Random buddies, a runner first probes its random buddy, only when none
     * of the random buddies responds it starts to probe its lifeline buddies
     */
    val victims:Rail[Long];

    /** Logger to record the work-stealing status */
    val logger:Logger;
    public var logManager:LogManager;
    public var processingStackManager:ProcessingStackManager;
    public var waitingAndWorkingManager:WaitingAndWorkingManager;
    //public var sleStateWaiting:SimpleLogEvent;
    public var sleTrySteal:SimpleLogEvent;
    public var sleProcessSteals:SimpleLogEvent;
    public var sleDistribute:SimpleLogEvent;
    public var sleBfAck:SimpleLogEvent;

    /* 
     * Number of places.
     */
    public var P:Long;

    /**
     * Manages the open-list.
     * The open-list is used when i am acting as a victim.
     */
    public var openManager:OpenManager[Queue, R];
    public var currentTan:Long;

    /**
     * Manages the openFEnd-list.
     * The openFEnd-list is used when i am acting as a thief.
     */
    public var openFEndManager:OpenFEndManager;

    /**
     * Manages sending backups to the backup-place (back).
     */
    public var receiver:BackupReceiverManager[Queue, R];

    /**
     * Manages receiving backups from the backedup-place (forth).
     */
    public var sender:BackupSenderManager[Queue, R];

    /**
     * Manages cyclic monitoring of places.
     */
    public var monitor:MonitorManager[Queue, R];

    /**
     * Manages the state of the worker. Everything regarding the state should
     * be saved in this manager.
     */
    public var state:StateManager;

    /**
     * Manages timeouts of places.
     */
    public var timeout:TimeoutManager;

    /**
     * Manages all recorded messages.
     */
    public var message:MessageManager[Queue, R];

    /**
     * Manages links to victim-places.
     * This is used, when i am acting as the backup of the thief.
     */
    public var linkManager:LinkManager;

    /**
     * This manager governs the id of the backup-place (back) and the backedup
     * place (forth) and updates them, if requested.
     */
    public var backForth:BackForthManager[Queue, R];

    /**
     * The PlaceLocalHandle to access other places.
     */
    public var st:PlaceLocalHandle[Worker[Queue, R]];

    /**
     * My place-id.
     */
    public val me:Long;

    /**
     * This attribute is alway in range [-w, z]. It is used to determine the
     * next place to send steal-requests to. If z == nextVictim, then all steal
     * were unsuccessful and the worker goes inactive.
     * This attribute is used when i am acting as a thief.
     */
    public var nextVictim:Long;
    public var nextRandom:Long;

    /**
     * This datastructure stores the higest tan-id it has ever received from
     * a victim in a give. We use a HashMap, since the number of places is not
     * fixed due to malleability.
     * The first Long is the place-id, the second Long is the highes received
     * tan to this place-id.
     * This attribute is used when i am acting as a thief.
     */
    public var gotTasksFrom:HashMap[Long, Long];

    /**
     * The tanDescriptor stores all places with their assigned loot for the
     * currently active transaction (steals).
     * This attribute is used when I am acting as a victim.
     */
    public var tanDescriptor:ArrayList[Pair[Long, TaskBag]];

    public val emergencyMerges:SimpleStack[TaskBag];

    private var lifelineList:HashMap[Long, ArrayList[Long]];

    /**
     * Class constructor
     * @param init function closure to init the local {@link TaskQueue}
     * @param n same to this.n
     * @param w same to this.w
     * @param m same to this.m
     * @param l power of lifeline graph
     * @param z base of lifeline graph
     * @param tree true if the workload is dynamically generated,
     *             false if the workload can be statically generated
     */
    public def this(val init:()=>Queue
                    , val n:Int
                    , val w:Int
                    , val l:Int
                    , val z:Int
                    , val m:Int
                    , val k:Int
                    , val r:Int
                    , val tree:Boolean) : Worker[Queue, R]
    {
        this.P = Place.numPlaces();
        this.openManager = new OpenManager[Queue, R]();
        this.openFEndManager = new OpenFEndManager();
        this.timeout = new TimeoutManager();
        this.linkManager = new LinkManager();
        this.gotTasksFrom = new HashMap[Long, Long]();
        this.tanDescriptor = new ArrayList[Pair[Long, TaskBag]]();
        this.message = new MessageManager[Queue, R]();
        this.emergencyMerges = new SimpleStack[TaskBag](P);
        this.n = n;
        this.w = w;
        this.m = m;
        this.k = k;
        this.l = l;
        this.z = z;
        this.currentK = 0n;
        this.r = r;
        this.currentR = 0n;
        // this.lifelines = new Rail[Long](z, -1);
        this.lifelines = new ArrayList[Long]();

        this.me = Runtime.hereLong();

        // lifelines
        if (0 == me)
        {
            this.lifelineList = calculateAllLifelines(this.P, this.z, l);
            AssertPrinter.println("[Worker " + here  + "]: buddy-list: " + lifelineList);
            this.lifelines = this.lifelineList.get(0);
        }
        else
        {
            this.lifelines = calculateLifelines(this.P, this.me, this.z, l);
        }
        AssertPrinter.println("[Worker " + here  + "]: lifeline-list: " + this.lifelines);
        this.victims = new Rail[Long](m, -1);
        var i:Long = 0;
        fillVictims();
        this.nextVictim = -w;
        this.nextRandom = 0;
        this.queue = init();
        // this.lifelinesActivated = new Rail[Boolean](P, this.z);

        if (tree)
        {
            // 1st wave
            // if ((3 * this.me + 1) < P)
            // {
            //     this.message.lifelineThieves.push(3 * this.me + 1);
            // }
            // if ((3 * this.me + 2) < P)
            // {
            //     this.message.lifelineThieves.push(3 * this.me + 2);
            // }
            // if ((3 * this.me + 3) < P)
            // {
            //     this.message.lifelineThieves.push(3 * this.me + 3);
            // }
            // if (this.me > 0)
            // {
            //     this.message.lifelinesActivated.add((this.me - 1) / 3);
            // }
            val predecessors:Rail[Long] = calculatePredecessors(l, z, me, P);
            for (val predecessor in predecessors)
            {
                if (-1 != predecessor)
                {
                    this.message.lifelineThieves.push(predecessor);
                }
            }

            for (val id in lifelines)
            {
                this.message.lifelinesActivated.add(id);
            }
        }
        this.logger = new Logger(true);
        this.processingStackManager = new ProcessingStackManager();
        this.waitingAndWorkingManager = new WaitingAndWorkingManager();

        this.logManager = new LogManager(this.me, this.logger.timeReference,
                                         this.processingStackManager,
                                         this.waitingAndWorkingManager);

        this.state = new StateManager(waitingAndWorkingManager);

        //Give refTime to SimpleLogEvents
        this.timeout.sleTimeout.setRefTime(this.logManager.getRefTime());
        //this.sleStateWaiting =
        //    new SimpleLogEvent("StateWaiting", this.logManager.getRefTime());
        this.sleTrySteal =
            new SimpleLogEvent("TrySteal", this.logManager.getRefTime());
        this.sleProcessSteals =
            new SimpleLogEvent("ProcessSteals", this.logManager.getRefTime());
        this.sleDistribute =
            new SimpleLogEvent("Distribute", this.logManager.getRefTime());
        this.sleBfAck =
            new SimpleLogEvent("BfAck", this.logManager.getRefTime());

        AssertPrinter.println("[Worker " + here + "]: leaving constructor.");
    }

    public def this(  val init:()=>Queue
                    , val n:Int
                    , val w:Int
                    , val l:Int
                    , val m:Int
                    , val k:Int
                    , val r:Int) : Worker[Queue, R]
    {
        this.P = Place.numPlaces();
        this.openManager = new OpenManager[Queue, R]();
        this.openFEndManager = new OpenFEndManager();
        this.timeout = new TimeoutManager();
        this.linkManager = new LinkManager();
        this.gotTasksFrom = new HashMap[Long, Long]();
        this.tanDescriptor = new ArrayList[Pair[Long, TaskBag]]();
        this.message = new MessageManager[Queue, R]();
        this.emergencyMerges = new SimpleStack[TaskBag](P);
        this.n = n;
        this.w = w;
        this.m = m;
        this.k = k;
        this.l = l;
        this.currentK = 0n;
        this.r = r;
        this.currentR = 0n;
        // this.lifelines = new Rail[Long](0);
        this.lifelines = new ArrayList[Long]();

        this.me = Runtime.hereLong();

        this.victims = new Rail[Long](m, -1);
        fillVictims();

        this.nextVictim = -w;
        this.nextRandom = 0;
        this.queue = init();

        AssertPrinter.println("[Worker " + here + "]: my victims = " + this.victims);
        AssertPrinter.println("[Worker " + here + "]: my buddies = " + this.lifelines);
        AssertPrinter.println("[Worker " + here + "]: lifelineThieves = " + this.message.lifelineThieves);
        this.logger = new Logger(false);
        this.processingStackManager = new ProcessingStackManager();
        this.waitingAndWorkingManager = new WaitingAndWorkingManager();
        this.logManager = new LogManager(this.me, this.logger.timeReference,
        this.processingStackManager,
        this.waitingAndWorkingManager);

        this.state = new StateManager(waitingAndWorkingManager);

        //Give refTime to SimpleLogEvents
        this.timeout.sleTimeout.setRefTime(this.logManager.getRefTime());
        //this.sleStateWaiting =
        //    new SimpleLogEvent("StateWaiting", this.logManager.getRefTime());
        this.sleTrySteal =
            new SimpleLogEvent("TrySteal", this.logManager.getRefTime());

        this.sleProcessSteals =
            new SimpleLogEvent("ProcessSteals", this.logManager.getRefTime());
        this.sleDistribute =
            new SimpleLogEvent("Distribute", this.logManager.getRefTime());
        this.sleBfAck =
            new SimpleLogEvent("BfAck", this.logManager.getRefTime());
    }

    @Inline static def calculateAllLifelines(val P:Long, val l:Long)
        : HashMap[Long, ArrayList[Long]]
    {
        var z:Long = 1;
        while ((l << z) < P)
        {
            ++z;
        }
        return (calculateAllLifelines(P, z, l));
    }

    @Inline static def calculateAllLifelines(  val P:Long
                                             , val z:Long
                                             , val l:Long)
        : HashMap[Long, ArrayList[Long]]
    {
        val lifelineList:HashMap[Long, ArrayList[Long]] = new HashMap[Long, ArrayList[Long]]();
        for (val v:Long in 0..P)
        {
            val lifelines:ArrayList[Long] = calculateLifelines(P, v, z, l);
            if (!lifelines.isEmpty())
            {
                lifelineList.put(v, lifelines);
            }
        }

        return (lifelineList);
    }

    @Inline static def calculateLifelines(val P:Long, var id:Long, val z:Long, val l:Long)
    {
        val lifelines:ArrayList[Long] = new ArrayList[Long]();
        var x:Long = 1;
        var t:Long = 0;
        for (var j:Long = 0; j < z; j++)
        {
            var v:Long = id;
            for (var i:Long = 1; i < l; i++)
            {
                v = v - v % (x * l) + (v + (x * l) - x) % (x * l);
                if (v < P)
                {
                    lifelines.add(v);
                    break;
                }
            }
            x *= l;
        }

        return (lifelines);
    }

    static def calculatePredecessors(val l:Long,val z:Int, val id:Long, val p:Long)
        : Rail[Long]
    {
        val dim:Long = z;
        val nodecoutPerEdge:Long = l;

        val predecessors:Rail[Long] = new Rail[Long](dim);
        var mathPower_nodecoutPerEdge_I:Long = 1;
        for (var i:Long = 0; i < dim; i++) {
            // Von Startknoten die Vektorlaenge im Wuerfel auf der Achse i ermitteln:
            var vectorLength:Long = (id / mathPower_nodecoutPerEdge_I) % nodecoutPerEdge;

            // Wenn es einen normalen Ueberlauf gab oder der gewaehlte Vorgaengerknoten nicht exisitiert:
            // Der Vorgaengerknoten wird, nur wenn es kein normaler Ueberlauf ist, in predecessors[i] gespeichert
            if (vectorLength + 1 == nodecoutPerEdge || (predecessors(i) = id + mathPower_nodecoutPerEdge_I) >= p) {
                // Standard Ueberkaufknoten waehlen (ist in beiden Faellen der selbe)
                predecessors(i) = id - (vectorLength * mathPower_nodecoutPerEdge_I);

                // Falls Ueberlaufknoten alleine in dieser Richtung ist und deshalb nun auf sich
                // selbst zeigt diesen ignorieren:
                if (predecessors(i) == id)
                {
                    predecessors(i) = -1;
                }
            }
            mathPower_nodecoutPerEdge_I *= nodecoutPerEdge;
        }
        return predecessors;
    }

    public def fillVictims() : void
    {
        if (P > 1)
        {
            var i:Long = 0;
            while (i < this.m)
            {
                val v = this.random.nextLong(P);
                if (v == this.me || Place(P).isDead())
                {
                    continue;
                }
                // for (val lifeline:Long in this.lifelines)
                // {
                //     if (v == lifeline)
                //     {
                //         continue;
                //     }
                // }
                // for (val victim:Long in this.victims)
                // {
                //     if (victim == v)
                //     {
                //         continue;
                //     }
                // }
                victims(i++) = v;
            }
        }
    }

    /**
     * Main process function of Worker.
     */
    final def processStack()
    {
        //Start debugging
        //var labelCurrLE:String;
        //End debugging
        //this.logManager.startLE(new LEworking(System.nanoTime()));
       
       AssertPrinter.println("[Worker " + here + "]: Starting.");
       while (this.queue.size() > 0 || !this.state.stealFailed)
       {
           //Start debugging.
           //labelCurrLE = logManager.currLE.label;
           //if(!(labelCurrLE.equals("working")))
           //{
           //    Console.OUT.println("Error: Falsches LE in processStack an Stelle 2: " + labelCurrLE);
           //}
           //End debugging.
           // mb: code to test crashes of a single place (Place(1)) at a random point in time.
//            if (   /*this.logger.timeReference + (5E9 as Long) - System.nanoTime() <= 0
//                && */ this.me == 1
//                && this.random.nextDouble() < 0.1d)
//            {
//                Console.OUT.println(  here + ": crashing me.");
// 
//                System.killHere();
//            }

           //     try
           //     {
           //         at (Place(3)) { System.killHere(); }
           //     }
           //     catch (val t:CheckedThrowable) { }

           //     System.killHere();
           // }
           // if (here.id() == this.P - 1)
           // {
           //     if (this.random.nextDouble() < 0.1)
           //     {
           //         System.killHere();
           //     }
           // }
           if (this.queue.size() > 0 && !this.state.emergency)
           {
               processingStackManager.startProcessing();
               this.queue.process(n);
               // for (i:Long in 0..(this.n - 1))
               // {
               //     beep = beep || this.state.emergency;
               //     this.queue.process(1);
               //     this.prot(i);
               //     if (0 == this.queue.size())
               //     {
               //         break;
               //     }
               // }
               processingStackManager.endProcessing();
           }

           if (0 == this.queue.size())
           {
               this.state.empty = true;
           }
           else
           {
               ++this.currentK;
           }
           Runtime.probe();
           if (0 == this.queue.size() && this.state.stealFailed)
           {
               if (Place(this.backForth.getBack()).isDead())
               {
                   this.isDead(this.backForth.getBack());
               }
           }
           //Start debugging.
           //labelCurrLE = logManager.currLE.label;
           //if(!(labelCurrLE.equals("working")))
           //{
           //    Console.OUT.println("Error: Falsches LE in processStack an Stelle 2: " + labelCurrLE);
           //}
           //End debugging.
           this.processRecorded();
       }
       //Start debugging.
       //labelCurrLE = logManager.currLE.label;
       //if(!(labelCurrLE.equals("working")))
       //{
       //   Console.OUT.println("Error: Falsches LE in processStack an Stelle 3: " + labelCurrLE);
       //}
       //End debugging.
       this.processRecorded();
       AssertPrinter.println("[Worker " + here + "]: Stopping. queueSize = " + this.queue.size() + ", recordedMerges = " + this.message.recordedMerges.size());
       AssertPrinter.println("[Worker " + here + "]: myResult (so far) = " + this.queue.getResult().getResult() + ".");
       
       //this.logManager.endLE();
       //this.logManager.startLE(new LEIdling(System.nanoTime()));
    }

    /**
     * Entry point when workload is only known dynamically . The workflow is
     * terminated when 
     * (1) No one has work to do
     * (2) Lifeline steals are responded
     * @param start init method used in {@link TaskQueue}, note the workload is
     * not allocated, because the workload can only be self-generated.
     */
    def main(start:() => void) : void
    {
        finish
        {
            val P:Long = Place.numPlaces();
            val st:PlaceLocalHandle[Worker[Queue, R]] = this.st;
            if (256 > P)
            {
                finish
                {
                    for (id:Long in 0..(P - 1))
                    {
                        val place:Place = Place(id);
                        at (place) async
                        {
                            // st().prot(0);
                            st().state.sendingBackup = true;
                            st().sender.sendFullBackupRegular(st().back());
                            while (st().state.sendingBackup)
                            {
                                Runtime.probe();
                            }

                            AssertPrinter.println(  "[Worker " + here
                                                  + "]: init Backup written.");
                        }
                    }
                }
            }
            else // if (256 <= P)
            {
                finish
                {
                    for (var i:Long = P - 1; i >= 0; i -= 32)
                    {
                        at (Place(i)) async
                        {
                            val max = Runtime.hereLong();
                            val min = Math.max(max - 31, 0);
                            for (var j:Long = min; j <= max; ++j)
                            {
                                at (Place(j)) async
                                {
                                    // st().prot(0);
                                    st().state.sendingBackup = true;

                                    st().sender
                                        .sendFullBackupRegular(st().back());

                                    while (st().state.sendingBackup)
                                    {
                                        Runtime.probe();
                                    }

                                    AssertPrinter.println(  "[Worker " + here
                                                          + "]: init Backup "
                                                          + "written.");
                                }
                            }
                        }
                    }
                }
            }
        }
        finish
        {
            try
            {
                this.state.empty = false;
                this.state.active = true;
                this.logger.startLive();
                this.logManager.startLE(new LEworking(System.nanoTime()));
                start();
                this.processStack();
                this.logger.stopLive();
                this.logManager.endLE();
                this.state.active = false;
                this.logger.nodesCount = this.queue.count();
            }
            catch (val v:CheckedThrowable)
            {
                error(v);
            }
        } 
    }

    /**
     * Entry point when workload can be known statically. The workflow is
     * terminated when:
     * (1) No one has work to do
     * (2) Lifeline steals are responded
     */
    def main()
    {
        try
        {
            this.state.empty = false;
            this.state.active = true;
            this.logger.startLive();
            this.logManager.startLE(new LEworking(System.nanoTime()));
            this.processStack();
            this.logger.stopLive();
            this.logManager.endLE();
            this.state.active = false;
            this.logger.nodesCount = this.queue.count();
        }
        catch (val v:CheckedThrowable)
        {
            error(v);
        }
    }

    /**
     * Print exceptions
     * @param v exeception
     */
    static def error(val v:CheckedThrowable) : void
    {
        Runtime.println("Exception at " + here);
        v.printStackTrace();
    }

    /**
     * Min helper function
     */
    @Inline static def min(i:Long, j:Long) : Long
    {
        return (i < j ? i : j);
    }

    /**
     * Internal method used by {@link GLB} to start Worker at each place when
     * the workload is known statically.
     * @param st PLH of Worker
     */
    static def broadcast[Queue, R](st:PlaceLocalHandle[Worker[Queue, R]])
        {Queue <: TaskQueue[Queue, R]} : void
    {
        val P = Place.numPlaces();
        finish
        {
            if (256 > P)
            {
                finish
                {
                    for(var i:Long = 0; i < P; i++)
                    {
                        at (Place(i)) async
                        {
                            // st().prot(0);
                            st().state.sendingBackup = true;

                            AssertPrinter.println(  "[Worker " + here
                                                  + "]: writing init Backup.");

                            st().sender.sendFullBackupRegular(st().back());

                            while (st().state.sendingBackup)
                            {
                                Runtime.probe();
                            }

                            AssertPrinter.println(  "[Worker " + here
                                                  + "]: init Backup written.");
                        }
                    }
                }
                for(var i:Long = 0; i < P; i++)
                {
                    at (Place(i)) async
                    {
                        st().main();
                    }
                }
            }
            else // if (256 <= P)
            {
                finish
                {
                    for (var i:Long= P - 1; i >= 0; i -= 32)
                    {
                        at (Place(i)) async
                        {
                            val max = Runtime.hereLong();
                            val min = Math.max(max - 31, 0);
                            for (var j:Long = min; j <= max; ++j)
                            {
                                at (Place(j)) async
                                {
                                    // st().prot(0);
                                    st().state.sendingBackup = true;

                                    st().sender
                                        .sendFullBackupRegular(st().back());

                                    while (st().state.sendingBackup)
                                    {
                                        Runtime.probe();
                                    }

                                    AssertPrinter.println(  "[Worker " + here
                                                          + "]: init Backup "
                                                          + "written.");
                                }
                            }
                        }
                    }
                }
                for(var i:Long = P - 1; i >= 0; i -= 32)
                {
                    at (Place(i)) async
                    {
                        val max = Runtime.hereLong();
                        val min = Math.max(max - 31, 0);
                        for (var j:Long = min; j <= max; ++j)
                        {
                            at (Place(j)) async
                            {
                                st().main();
                            }
                        }
                    }
                }
            }
        }
    }

    /**
     * Initialize the placeLocalHandle as well as some managers on each place.
     * @param st: PLH of Worker
     */
    static def initContexts[Queue,R](val st:PlaceLocalHandle[Worker[Queue, R]])
        {Queue<:TaskQueue[Queue, R]} : void
    {
        val P = Place.numPlaces();
        val lifelineList:HashMap[Long, ArrayList[Long]] = st().lifelineList;
        finish
        {
            if (P < 256)
            {
                for(var i:Long = 0; i < P; i++)
                {
                    val lifelines:ArrayList[Long] = lifelineList.get(i);
                    assert(null != lifelines);
                    at (Place(i)) async
                    {
                        st().init(st);
                        st().lifelines = lifelines;
                    }
                }
            }
            else
            {
                for(var i:Long = P - 1; i >= 0; i -= 32)
                {
                    val max = i;
                    val min = Math.max(max - 31, 0);

                    val localLifelineList:HashMap[Long, ArrayList[Long]]
                        = new HashMap[Long, ArrayList[Long]]();

                    for (var j:Long = min; j <= max; ++j)
                    {
                        val list:ArrayList[Long] = lifelineList.get(j);
                        assert (null != list);
                        localLifelineList.put(j, list);
                    }

                    at (Place(i)) async
                    {
                        for (var j:Long = min; j <= max; ++j)
                        {
                            val lifelines:ArrayList[Long] = localLifelineList.get(j);
                            assert (null != lifelines);
                            at (Place(j)) async
                            {
                                st().init(st);
                                st().lifelines = lifelines;
                            }
                        }
                    }
                }
            }
        }
    }

    // New stuff
    /**
     * Method to initialize the placeLocalHandle as well as managers, which
     * need a reference to the this-object.
     */
    protected def init(val st:PlaceLocalHandle[Worker[Queue, R]])
    {
        this.st = st;
        this.backForth = new BackForthManager(this);
        this.receiver = new BackupReceiverManager[Queue, R](this);
        this.sender = new BackupSenderManager[Queue, R](this);
        this.monitor = new MonitorManager[Queue, R](this);
        this.message.worker = this;
        AssertPrinter.println("[Worker " + here + "]: my victims = " + this.victims);
        AssertPrinter.println("[Worker " + here + "]: my buddies = " + this.lifelines);
        AssertPrinter.println("[Worker " + here + "]: lifelineThieves = " + this.message.lifelineThieves);
        AssertPrinter.println("[Worker " + here + "]: backup-place = " + this.backForth.getBack());
        AssertPrinter.println("[Worker " + here + ": init done.");
    }

    @Inline private def resetNextVictim()
    {
        this.nextVictim = -this.w;
    }

    /**
     * Request a regular backup.
     * This is a general-purpose method and can be called regardless of the
     * role i am fulfilling in the moment (victim, thief, back(victim), 
     * back(thief)).
     */
    public def regReq() : void
    {
        assert (this.state.canSendBackup())
               : here + ": trying to send a backup while already sending a "
                      + "backup!";

        val backupId:Long = this.back();

        val t:Timeout = new Timeout(backupId
                                    , Timeout.REG_ACK
                                    , Worker.TIMEOUT_REGULAR);

        if (!this.timeout.add(t))
        {
            this.isDead(backupId);
            return;
        }
        ++this.logger.regularBackupsWritten;
        this.state.sendingBackup = true;
        // if (this.state.afterMerge)
        // {
            this.sender.sendFullBackupRegular(backupId);
        // }
        // else // if (!this.state.afterMerge)
        // {
        //     this.sender.sendIncrementalBackupRegular(backupId);
        // }
    }

    /**
     * Process all recorded messages in order.
     * This is a general-purpose method and can be called regardless of the
     * role i am fulfilling in the moment (victim, thief, back(victim), 
     * back(thief)).
     */
    public def processRecorded() : void
    {
        ++this.currentR;
        var skipRegular:Boolean = false;
        this.processRestores();
        this.processMerges();
        if (this.state.sendBackupTO)
        {
            val backupId:Long = this.backForth.getBack();

            val timeout:Timeout
                = new Timeout(backupId, Timeout.TO_ACK, Worker.TIMEOUT_TO);

            if (!this.timeout.add(timeout))
            {
                this.isDead(backupId);
            }
            else if (this.state.canSendBackup())
            {
                this.state.sendBackupTO = false;
                this.state.sendingBackup = true;
                this.sender.sendFullBackupTO(backupId);
            }
        }
        //In give wird vom thief state.empty = false gesetzt.
        if (this.state.empty)
        {
            this.noTasks();
            if (   !this.state.waiting
                && this.state.active
                && (1 > this.tanDescriptor.size()))
            {
                this.trySteal();
            }
        }
        else if (this.state.canSendBackup() && this.tanDescriptor.isEmpty())
        {
            skipRegular = this.processSteals();
        }

        if (  (this.k <= this.currentK)
            && this.state.canSendBackup()
            && !skipRegular)
        {
            this.regReq();
            this.currentR = 0n;
        }
        else if (this.r <= this.currentR)
        {
            this.monitor();
            if (0 == this.me)
            {
                this.checkJoins();
            }
        }
        this.processTimeouts();
    }

    public def processTimeouts() : void
    {
        val timeouts:List[Timeout] = this.timeout.check();
        if (null != timeouts)
        {
            AssertPrinter.println("[Worker " + here + "]: " + timeouts);
            for (timeout:Timeout in timeouts)
            {
                // A victim died -> ignore it, go to the next victim
                if ((Timeout.GIVE & timeout.reason) > 0)
                {
                    ++this.nextVictim;
                    //if (true == this.state.waiting)
                    //{
                    //    this.state.waiting = false;
                    //    this.sleStateWaiting.setET();
                    //} else {
                    //    this.state.waiting = false;
                    //}
                    this.state.setWaiting(false);
                    AssertPrinter.println(  "[Worker " + here + "]: "
                                          + "(in processTimeouts): empty = "
                                          + this.state.empty
                                          + ", waiting = " + this.state.waiting
                                          + ", nextVictim = " + this.nextVictim
                                          + ", tanDSize = " 
                                          + this.tanDescriptor.size()
                                          + ", queueSize = "
                                          + this.queue.size());
                }
                // A place which has to answer to a gotChedk or linkResolve
                // died -> crashing program.
                else if ((    (  Timeout.GOT_ACK
                               | Timeout.LNK_RES)
                           & timeout.reason) > 0)
                {
                    Console.OUT.println(  "A place I sent a gotCheck or a "
                                        + " linkResolve died. Crashing"
                                        + " program.");

                    Worker.killAll();
                    Runtime.probe(); // To receive the killAll
                }
                // My backup-place died within a regular backup
                // -> remove timeout, communicate dead place.
                else if ((Timeout.REG_ACK & timeout.reason) > 0)
                {
                    this.state.sendingBackup = false;
                }
                // My backup-place died within a steal-backup
                // -> write IA-Backup as complete backup and
                //    set repeatStlReq to "true" (IAAck will also act as
                //    stlAck).
                else if ((Timeout.STL_ACK & timeout.reason) > 0)
                {
                    this.state.sendingBackup = false;
                    this.state.repeatStlReq = true;
                }
                // A thief of mine died -> re-merge its loot into own queue.
                else if ((Timeout.BF_ACK & timeout.reason) > 0)
                {
                    val entry:Pair[Long, TaskBag]
                        = this.openManager.get(this.currentTan
                                               , timeout.placeId);

                    val bag:TaskBag = entry.second;
                    if (null != bag)
                    {
                        this.openManager.remove(this.currentTan
                                                , timeout.placeId);

                        this.queue.merge(bag);
                    }
                    this.bfAck(this.currentTan, timeout.placeId, false);
                }
                else if ((Timeout.RST_ACK & timeout.reason) > 0)
                {
                    this.state.answerRstAckAfterIAAck = true;
                }
                else if ((Timeout.PROCESS_JOINS & timeout.reason) > 0)
                {
                    this.processJoins();
                }

                if (this.backForth.getBack() == timeout.placeId)
                {
                    this.isDead(timeout.placeId);
                }
            }
        }
    }
    /**
     * Request a steal at the next possible victim place.
     * To know who the next possible victim is, this method needs to calculate
     * its (victims) id.
     * This method is called, when i am fulfilling the role as a thief. To be
     * precise: by calling this method, i become a thief.
     * This method is not a sink-method. It has place-changes.
     * Role: thief
     * sink: no
     */
    public def trySteal() : void
    {
        this.sleTrySteal.setST();
        if (   Place(this.backForth.getBack()).isDead())
        {
            this.isDead(this.backForth.getBack());
            this.sleTrySteal.setET();
            return;
        }

        if (   this.state.waiting
            || (0 == Place.numPlaces())
            || this.state.emergency)
        {
            this.sleTrySteal.setET();
            return;
        }

        this.state.empty = true;
        if (1 == P)
        {
            this.state.stealFailed = true;
            this.sleTrySteal.setET();
            return;
        }

        var succeeded:Boolean = false;
        while (!succeeded)
        {
            var leTrySteal:LEtrySteal = new LEtrySteal(System.nanoTime());
            this.logManager.startLE(leTrySteal);
            var stealId:Long = this.me;
            val victimId = this.getNextVictimId();
            leTrySteal.victimId = victimId;

            AssertPrinter.println(  "[Worker " + here + "]: next victim-id = "
                                  + victimId);

            if (victimId < 0) // No new victim found
            {
                assert (this.lifelines.size() <= this.nextVictim)
                       : here + ": still victims left!\n"
                        + this.nextVictim;

                this.state.stealFailed = true;
                this.logManager.endLE();
                this.sleTrySteal.setET();
                return;
            }
            else if (0 <= this.nextVictim) // lifeline-buddy -> activate lifeline
            {
                assert (!this.message.isLifelineActive(victimId))
                       : here + ": cannot activate a lifeline twice.";

                this.message.lifelinesActivated.add(victimId);
                ++this.logger.lifelineStealsAttempted;
            }
            else // if (0 > this.nextVictim) // random steal
            {
                stealId = -stealId - 1;
                ++this.logger.stealsAttempted;
            }

            val me = stealId;
            leTrySteal.setStealingType(me);
            val victim:Place = Place(victimId);
            val timeout:Timeout
                = new Timeout(victimId, Timeout.GIVE, Worker.TIMEOUT_GIVE);

            if (!this.timeout.add(timeout))
            {
                continue;
            }
            /*if (false == this.state.waiting)
            {
                this.state.waiting = true;
                this.sleStateWaiting.setST();
            } else {
                this.state.waiting = true;
                }*/
            this.state.setWaiting(true);
            val st:PlaceLocalHandle[Worker[Queue, R]] = this.st;

            AssertPrinter.println(  "[Worker " + here + "]: trying to steal at "
                                  + victim + " as "
                                  + (me < 0 ? "random":"buddy") + ".");
            
            leTrySteal.wasSent = true;
            leTrySteal.sendingTime = System.nanoTime();
            at (victim) @Uncounted async
            {
                st().pushThief(me);
            }
            succeeded = true;
            this.logManager.endLE();
        }
        this.sleTrySteal.setET();
    }

    public def pushThief(val id:Long) : void
    {
        var leStealReceiving:LEstealReceiving = new LEstealReceiving(System.nanoTime());
        this.logManager.startLE(leStealReceiving);
        leStealReceiving.setThiefId(id);
        leStealReceiving.setStealingType(id);
        if (0 > id)
        {
            ++this.logger.stealsReceived;
        }
        else // if (0 <= id)
        {
            ++this.logger.lifelineStealsReceived;
        }

        if (   !this.state.empty
            && this.state.active
            && (0 == this.tanDescriptor.size()))
        {
            AssertPrinter.println(  "[Worker " + here + "]: pushing " + id
                                  + " into thief-list. "
                                  + " empty = " + this.state.empty
                                  + " active = " + this.state.active
                                  + " tansOpen = " + this.tanDescriptor.size());
            this.message.thieves.push(id);
            leStealReceiving.setOnThieves();
        }
        else // if (   this.state.empty 
             //     || !this.state.active
             //     || 0 < this.tanDescriptor.size())
        {
            //var leNoTasks:LEnoTasks = new LEnoTasks(System.nanoTime());
            //this.logManager.startLE(leNoTasks);
            if (id >= 0)
            {
                //leNoTasks.lifelineSteal = true;
                this.message.lifelineThieves.push(id);
                leStealReceiving.setOnLifelineThieves();
            }

            val me:Long = this.me;
            val thief:Place = Place((0 > id) ? -id - 1 : id);
            val st:PlaceLocalHandle[Worker[Queue, R]] = this.st;

            AssertPrinter.println(  "[Worker " + here + "]: un-waiting " 
                                  + thief + ".");

            //leNoTasks.thiefId = ((0 > id) ? -id - 1 : id);
            at (thief) @Uncounted async
            {
                var leStealReject:LEstealReject = new LEstealReject(System.nanoTime());
                st().logManager.startLE(leStealReject);
                leStealReject.victimId = me;
                leStealReject.rejectType = 0; //Rejected in pushThief.
                ++st().nextVictim;
                /*if (true == st().state.waiting)
                {
                    st().state.waiting = false;
                    st().sleStateWaiting.setET();
                }
                else
                {
                    st().state.waiting = false;
                    }*/
                st().state.setWaiting(false);
                /*if (st().logManager.currLE.label.equals("idling"))
                {
                    Console.OUT.println("Debug: Call endWaiting in pushThief.");
                    st().logManager.currLE.endWaiting(System.nanoTime());
                    }*/
                st().timeout.remove(me, Timeout.GIVE);
                st().logManager.endLE();
            }
            //this.logManager.endLE();
        }
        this.logManager.endLE();
    }

    /**
     * Calculates the id of the next possible victim (lifeline or not) and
     * returns is.
     * Role: thief
     * Sink: yes
     */
    private def getNextVictimId() : Long
    {
        var victimId:Long = -1;
        if (this.nextVictim < 0) // Random steal
        {
            victimId = this.victims(this.nextRandom++);
            this.nextRandom %= this.m;
            return (victimId);
        }
        else // if (this.nextVictim >= 0) // lifeline-steal
        {
            for (;this.nextVictim < this.lifelines.size(); ++this.nextVictim)
            {
                victimId = this.lifelines.get(this.nextVictim);
                if (!this.message.lifelinesActivated.contains(victimId))
                {   // Unactivated lifeline -> we can use this!
                    return (victimId);
                }
            }
        }
        return (-1);
    }

    /**
     * Process all recorded steals.
     * Role: victim
     * Sink: no, changing place indirecty via sendStealBackup.
     */
    public def processSteals() : Boolean
    {
        this.sleProcessSteals.setST();
        //var leProcessSteals:LEprocessSteals
        //    = new LEprocessSteals(System.nanoTime());
        //this.logManager.startLE(leProcessSteals);
        if (this.state.empty)
        {
            this.noTasks();
            //this.logManager.endLE();
            this.sleProcessSteals.setET();
            return (false);
        }
        assert (this.tanDescriptor.isEmpty())
               : here
                 + ": tanDescriptor is not empty, but processSteal was called.";

        if (!(   this.message.thieves.isEmpty()
              && this.message.lifelineThieves.isEmpty()))
        {
            AssertPrinter.println(  "[Worker " + here + "]: thieves = "
                                  + this.message.thieves);

            AssertPrinter.println(  "[Worker " + here + "]: lifelineThieves = "
                                  + this.message.lifelineThieves);
        }
        val thieves:List[Long] = new ArrayList[Long]();
        var nStolen:Long = this.distribute(thieves);
        this.logger.nodesGiven += nStolen;

        this.noTasks();
        if (0 < this.tanDescriptor.size())
        {
            AssertPrinter.println(  "[Worker " + here + "]: all thieves = "
                                  + thieves);

            this.state.sendingBackup = true;
            this.currentTan = openManager.add(this.tanDescriptor);
            val backupId:Long = this.back();

            val timeout:Timeout = new Timeout(backupId
                                              , Timeout.STL_ACK
                                              , Worker.TIMEOUT_STEAL);

            if (!this.timeout.add(timeout))
            {
                this.isDead(backupId);
            }

            ++this.logger.stealBackupsWritten;
            this.sender.sendStealBackup(this.currentTan
                                        , backupId
                                        , nStolen
                                        , thieves);

            //this.logManager.endLE();
            this.sleProcessSteals.setET();
            return (true);
        }
        //this.logManager.endLE();
        this.sleProcessSteals.setET();
        return (false);
    }

    /**
     * Assign lootbags to thieves and store it in the tanDescriptor. Return
     * count of stolen tasks (as long).
     * Role: victim
     * Sink: yes
     */
    private def distribute(val thieves:List[Long]) : Long
    {
        //Start debugging.
        val labelCurrLE:String = logManager.currLE.label;
        //if(!(labelCurrLE.equals("working") || labelCurrLE.equals("idling")
        //     || labelCurrLE.equals("receiveBackup"))
        //{
        //    
        //    Console.OUT.println("Error: Falsches LE in distribute auf Place "
        //                        + this.me  + " an Stelle 4: " + labelCurrLE);
        //    Console.OUT.println(logManager.currLE.toString(0));
        //    
        //}
        //End debugging.
        this.waitingAndWorkingManager.startDistributing();
        this.sleDistribute.setST();
        //var leDistribute:LEdistribute = new LEdistribute(System.nanoTime());
        //this.logManager.startLE(leDistribute);
        
        var nStolen:Long = 0;
        // val oldHash:Long = this.queue.hash(); // mb: for debugging
        // val hashList:ArrayList[Long] = new ArrayList[Long]();
        while (!this.message.thieves.isEmpty())
        {
            var thiefId:Long = this.message.thieves.peek();
            if (thiefId < 0)
            {
                thiefId = -thiefId - 1;
                ++this.logger.stealsSuffered;
            }
            else
            {
                ++this.logger.lifelineStealsSuffered;
            }

            if (Place(thiefId).isDead())
            {
                continue;
            }
            val bag:TaskBag = this.queue.split();
            if (null == bag || 1 > bag.size())
            {
                //this.logManager.endLE();
                this.waitingAndWorkingManager.endDistributing();
                this.sleDistribute.setET();
                // mb: for debugging
                // var hashSum:Long = 0;
                // for (val hash:Long in hashList)
                // {
                //     hashSum += hash;
                // }
                // hashSum += this.queue.hash();

                // AssertPrinter.println(  "[Worker " + here + "]: oldHash = " + oldHash
                //                       + ", newHash = " + hashSum + ", equals = "
                //                       + (oldHash == hashSum));

                return (nStolen);
            }
            this.message.thieves.pop();
            // hashList.add(bag.hash());
            nStolen += bag.size();

            val entry:Pair[Long, TaskBag]
                = new Pair[Long, TaskBag](thiefId, bag);

            if (thieves.contains(thiefId))
            {
                for (val existing:Pair[Long, TaskBag] in this.tanDescriptor)
                {
                    if (existing.first == thiefId)
                    {
                        existing.second.merge(entry.second);
                    }
                }
            }
            else
            {
                thieves.add(thiefId);
                this.tanDescriptor.add(entry);
            }
        }
        while (!this.message.lifelineThieves.isEmpty())
        {
            val thiefId:Long = this.message.lifelineThieves.peek();
            if (Place(thiefId).isDead())
            {
                continue;
            }

            val bag:TaskBag = this.queue.split();
            if (null == bag || 1 > bag.size())
            {
                // mb: for debugging
                // var hashSum:Long = 0;
                // for (val hash:Long in hashList)
                // {
                //     hashSum += hash;
                // }
                // hashSum += this.queue.hash();

                // AssertPrinter.println(  "[Worker " + here + "]: oldHash = " + oldHash
                //                       + ", newHash = " + hashSum + ", equals = "
                //                       + (oldHash == hashSum));

                //this.logManager.endLE();
                this.waitingAndWorkingManager.endDistributing();
                this.sleDistribute.setET();
                return (nStolen);
            }
            this.message.lifelineThieves.pop();
            // hashList.add(bag.hash());
            nStolen += bag.size();

            val entry:Pair[Long, TaskBag]
                = new Pair[Long, TaskBag](thiefId, bag);

            if (thieves.contains(thiefId))
            {
                for (val existing:Pair[Long, TaskBag] in this.tanDescriptor)
                {
                    if (existing.first == thiefId)
                    {
                        existing.second.merge(entry.second);
                    }
                }
            }
            else if (thieves.contains(-thiefId-1))
            {
                thieves.remove(-thiefId-1);
                thieves.add(thiefId);
                for (var idx:Long = 0; idx < this.tanDescriptor.size(); ++idx)
                {
                    val existing:Pair[Long, TaskBag]
                        = this.tanDescriptor.get(idx);

                    if (existing.first == thiefId)
                    {
                        this.tanDescriptor.removeAt(idx);
                        entry.second.merge(existing.second);
                        this.tanDescriptor.add(entry);
                        break;
                    }
                }
            }
            else
            {
                thieves.add(thiefId);
                this.tanDescriptor.add(entry);
            }
        }
        if (!this.message.lifelineThieves.isEmpty())
        {
            AssertPrinter.println(  "[Worker " + here 
                                  + "]: unsatisfied lifeline-thieves = "
                                  + this.message.lifelineThieves);
        }
        // mb: for debugging
//         if (nStolen > 0)
//         {
//             var hashSum:Long = 0;
//             for (val hash:Long in hashList)
//             {
//                 hashSum += hash;
//             }
//             hashSum += this.queue.hash();
// 
//             AssertPrinter.println(  "[Worker " + here + "]: oldHash = " + oldHash
//                                   + ", newHash = " + hashSum + ", equals = "
//                                   + (oldHash == hashSum));
//         }

        //this.logManager.endLE();
        this.waitingAndWorkingManager.endDistributing();
        this.sleDistribute.setET();
        return (nStolen);
    }


    /**
     * 
     */
    @Inline public def give(val tan:Long
                            , val victimId:Long
                            , val loot:TaskBag) : void
    {
        this.timeout.remove(me, Timeout.GIVE);
        this.give(tan, victimId, loot, false);
    }

    /**
     * sends loot to thieves after steal-backup is written.
     * Role: victim
     * Sink: no, change place to thieves to give loot.
     */
    private def give(val tan:Long
                     , val victimId:Long
                     , val loot:TaskBag
                     , val isRecorded:Boolean) : Boolean
    {
        var leGive:LEgive = new LEgive(System.nanoTime());
        this.logManager.startLE(leGive);
        leGive.tan = tan;
        leGive.lootSize = loot.size();
        leGive.isRecorded = isRecorded;
        leGive.victimId = victimId;
        if (!this.state.sendingBackup)
        {
            leGive.merging = true;
            val backup:Place = Place(this.backForth.getBack());
            val me = this.me;
            val st:PlaceLocalHandle[Worker[Queue, R]] = this.st;

            AssertPrinter.println(  "[Worker "+ here + "]: sending link to "
                                  + backup);

            at (backup) async
            {
                st().victimLink(tan, victimId, me);
            }

            if (   !this.gotTasksFrom.containsKey(victimId)
                || this.gotTasksFrom.get(victimId) < tan)
            {
                this.gotTasksFrom.put(victimId, tan);
            }

            // Reset lifeline, if it was activated
            if (this.message.lifelinesActivated.contains(victimId))
            {
                this.message.lifelinesActivated.remove(victimId);
                this.logger.lifelineNodesReceived += loot.size();
                ++this.logger.lifelineStealsPerpetrated;
            }
            else
            {
                this.logger.nodesReceived += loot.size();
                ++this.logger.stealsPerpetrated;
            }

            val victim:Place = Place(victimId);

            AssertPrinter.println(  "[Worker " + here + "]: got " + loot.size()
                                  + " tasks from " + victim);

            this.openFEndManager.add(tan, victimId);
            this.resetNextVictim();
            /*if (true == this.state.waiting)
            {
                this.state.waiting = false;
                this.sleStateWaiting.setET();
            } else {
                this.state.waiting = false;
                }*/
            this.state.setWaiting(false);
            this.state.empty = false;
            if (!this.state.active)
            {

                //Console.OUT.println("Info: processStack() in give at Place " + this.me + ".");
                
                this.state.active = true;
                this.state.stealFailed = false;
                this.logger.startLive();
                this.logManager.startLE(new LEworking(System.nanoTime()));
                this.queue.merge(loot);
                // this.sender.prot();

                AssertPrinter.println(  "[Worker " + here + "]: after merge: queueSize = "
                                      + this.queue.size());

                this.state.afterMerge = true;
                this.processStack();
                this.state.active = false;
                this.logger.stopLive();
                this.logManager.endLE();
                this.logger.nodesCount = this.queue.count();
            }
            else
            {
                this.state.stealFailed = false;
                this.queue.merge(loot);
                // this.sender.prot();

                AssertPrinter.println(  "[Worker " + here + "]: after merge: queueSize = "
                                      + this.queue.size());

                this.state.afterMerge = true;
            }
        }
        else
        {
            if (isRecorded)
            {
                this.logManager.endLE();
                return (false);
            }
            val reason:String
                = (this.tanDescriptorContains(victimId) && victimId < this.me)
                  ? "Crossover"
                  : "Sending Backup";

            AssertPrinter.println(  "[Worker " + here
                                  + "]: merge recorded. Reason:" + reason);

            this.message.recordMerge(tan, victimId, loot);
        }
        this.logManager.endLE();
        return (true);
    }

    private def tanDescriptorContains(val id:Long) : Boolean
    {
        for (entry:Pair[Long, TaskBag] in this.tanDescriptor)
        {
            if (id == entry.first)
            {
                return (true);
            }
        }
        return (false);
    }

    /**
     * Link the tan to the current place. In case forth dies, we can pull
     * the stolen tasks from the victim.
     * Role: back(thief)
     * Sink: no, we need to send bfAck to the victim.
     */
    @Inline public def victimLink(val tan:Long
                                  , val victimId:Long
                                  , val thiefId:Long) : void
    {
        var leVictimLink:LEvictimLink = new LEvictimLink(System.nanoTime());
        this.logManager.startLE(leVictimLink);
        if (thiefId == this.forth()) 
        {
            this.linkManager.add(tan, victimId);
            val victim:Place = Place(victimId);
            val st:PlaceLocalHandle[Worker[Queue,R]] = this.st;

            AssertPrinter.println(  "[Worker " + here + "]: sending bfAck to "
                                  + victim + ".");

            at (victim) async
            {
                st().bfAck(tan, thiefId);
            }
        }
        else
        {
            assert (false) : here + "PANIC!";
        }
        this.logManager.endLE();
    }

    /**
     * Send negative steal-answers, if we cannot split our queue any further.
     * Also, increase every thieves's nextVictim by one.
     *
     * NOTE: every place can have only one victm-place it is waiting for. Since
     * thieves write themself ALWAYS in the thief-list and the victim extracts
     * lifelin-thieves at the first pass of the thief-list (moving them into
     * the lifeline-thieves list), this is uncritical.
     *
     * Role: victim
     * Sink: no, we have to change to thieves.
     */
    @Inline public def noTasks() : void
    {
        val st:PlaceLocalHandle[Worker[Queue, R]] = this.st;
        while (!this.message.thieves.isEmpty())
        {
            //var leNoTasks:LEnoTasks = new LEnoTasks(System.nanoTime());
            //this.logManager.startLE(leNoTasks);
            var thiefId:Long = this.message.thieves.pop();
            if (thiefId >= 0) // Lifeline-thief, store in lifeline-queue
            {
                //leNoTasks.lifelineSteal = true;
                this.message.lifelineThieves.push(thiefId);
            }
            else // if (thiefId < 0)
            {
                thiefId = -thiefId -1;
            }

            val thief:Place = Place(thiefId);
            val me:Long = this.me;
            //leNoTasks.thiefId = thiefId;
            at (thief) @Uncounted async
            {
                var leStealReject:LEstealReject = new LEstealReject(System.nanoTime());
                st().logManager.startLE(leStealReject);
                leStealReject.rejectType = me;
                leStealReject.rejectType = 1; //Rejected not in pushThief.
                ++st().nextVictim;
                /*if (true == st().state.waiting)
                {
                    st().state.waiting = false;
                    st().sleStateWaiting.setET();
                } else {
                    st().state.waiting = false;
                    }*/
                st().state.setWaiting(false);
                /*if (st().logManager.currLE.label.equals("idling"))
                {
                    Console.OUT.println("Debug: Call endWaiting in noTasks.");
                    st().logManager.currLE.endWaiting(System.nanoTime());
                }*/
                st().logManager.endLE();
            }
            //this.logManager.endLE();
        }
    }

    /**
     * TODO: What, if we are already sending a backup?
     */
    @Inline public def toReq() : void
    {
        val backupId:Long = this.back();
        this.sender.sendFullBackupTO(backupId);
    }

    /**
     * Our backup-place died. We need to inaugurate our new backup-place.
     * We take the backup-place of our backup-place as our new backup place.
     * Short: our new backup is back(back(me)).
     * Role: back(diedPlace)
     * Sink: no, we change the place (indirectly) via sendFullBackupIA.
     * TODO: unfinished?
     */
    @Inline public def iaReq() : void
    {
        val backupId:Long = this.back();

        val timeout:Timeout = new Timeout(backupId
                                          , Timeout.IA_ACK
                                          , Worker.TIMEOUT_RST);

        if (!this.timeout.add(timeout))
        {
            this.isDead(backupId);
            return;
        }
        this.state.sendingBackup = true;
        // if (this.state.answerRstAckAfterIAAck || this.state.repeatStlReq)
        // {
        //     this.sender.prot();
        // }
        this.sender.sendFullBackupIA(backupId);
    }

    /**
     * We are requested to restore a died place.
     * Role: back(diedPlace)
     * Sink: yes, message gets recorded.
     */
    @Inline public def restore(val tan:Long
                               , val diedPlace:Long
                               , val srcId:Long
                               , val thieves:List[Long]) : void
    {
        AssertPrinter.println(  "[Worker " + here + "]: recording Restore. "
                              + Place(diedPlace) + " died. " + Place(srcId)
                              + " told me.");

        this.message.recordRestore(tan, diedPlace, srcId, thieves);
        if (!this.state.active)
        {
            this.processRestores();
        }
    }

   /**
    * We need process a restore to back(diedPlace).
    * Role: forth(diedPlace)
    * Sink: no.
    */
    @Inline public def processRestores() : void
    {
        // val numRestores = this.message.recordedRestores.size();
        // AssertPrinter.println(here + ": numRestores = " + numRestores);
        while (!this.message.recordedRestores.isEmpty())
        {
            val restore:Restore = this.message.getRestore();
            var doNotWait:Boolean = true;
            val diedPlaceId:Long = restore.diedPlaceId;
            val srcId:Long = restore.srcId;
            val oldForthId:Long = this.forth();

            // Place was restored earlier -> nothing to do
            if (Worker.checkLog(me, this.sender.log, diedPlaceId))
            {
                val src:Place = Place(srcId);
                val st:PlaceLocalHandle[Worker[Queue, R]] = this.st;
                val me = this.me;

                assert (srcId == oldForthId)
                       :  "[Worker " + here  + "]: place is already restored, "
                        + "but my forth changed?! "
                        + srcId + " != " + oldForthId;

                at (src) async
                {
                    st().rstAck(me);
                }
                return;
            }
            // I am already fullfilling the role of back(deadPlace) in a
            // restore, a second restore must crash the program.
            else if (   this.timeout.containsReason(Timeout.GOT_ACK)
                     || this.timeout.containsReason(Timeout.LNK_RES)
                     || this.timeout.containsReason(Timeout.TO_ACK))
            {
                Console.OUT.println(  "[Worker " + here + "]: i have two "
                                    + "restores at the same time. Crashing "
                                    + "program.");

                Worker.killAll();
                Runtime.probe(); // to receive the killAll
            }

            this.backForth.updateForth();
            val forthId:Long = this.forth();

            assert (this.emergencyMerges.isEmpty())
                   :  here + ": starting restore, but have " 
                    + this.emergencyMerges.size() + " old emergency-merges "
                    + "left.";

            // The place that sent the restore-request has died before it was
            // able to write an IA-Backup. Therefore, data is definitely lost
            // -> crashing program
            if (srcId != forthId)
            {
                Console.OUT.println(  "[Worker " + here + "]: "
                                    + srcId + " != " + forthId
                                    + ". Crashing program.");

                Worker.killAll();
                Runtime.probe(); // To receive the killAll
                return;
            }

            assert (-1 == this.state.answerRestore) :  here
                                                     + ": answerRestore is "
                                                     + this.state.answerRestore
                                                     + ", but should be -1.";

            this.state.answerRestore = srcId;
            val me:Long = this.me;
            val st:PlaceLocalHandle[Worker[Queue, R]] = this.st;
            this.state.emergency = true;
            for (victimId:Long in this.linkManager.links.keySet())
            {
                val tans:ArrayList[Long]
                    = this.linkManager.links.get(victimId);

                if (!tans.isEmpty())
                {
                    doNotWait = false;
                    val victim:Place = Place(victimId);

                    val timeout:Timeout = new Timeout(victimId
                                                      , Timeout.LNK_RES
                                                      , Worker.TIMEOUT_LNK);

                    if (!this.timeout.add(timeout))
                    {
                        Console.OUT.println(  "One of my victims died as well."
                                            + "Crashing program.");

                        Worker.killAll();
                        Runtime.probe(); // To receive the killAll
                    }

                    AssertPrinter.println(  "[Worker " + here + "]: sending "
                                          + "resolve to " + victim + ".");

                    at (victim) async
                    {
                        st().linkResolve(oldForthId, me, tans);
                    }
                }
            }

            val tan:Long = this.receiver.tan;
            if (Long.MAX_VALUE > tan)
            {
                doNotWait = false;
                for (thiefId:Long in this.receiver.thieves)
                {
                    val timeout:Timeout = new Timeout(thiefId
                                                      , Timeout.GOT_ACK
                                                      , Worker.TIMEOUT_GOT);

                    if (!this.timeout.add(timeout))
                    {
                        Console.OUT.println(  "One of my thieves died as well."
                                            + "Crashing program.");

                        Worker.killAll();
                        Runtime.probe(); // To receive the killAll
                    }
                    val thief:Place = Place(thiefId);
                    at (thief) async
                    {
                        st().gotCheck(tan, diedPlaceId, me);
                    }
                }
            }

            // Check log
            if (!Worker.checkLog(oldForthId
                                  , this.receiver.backupLog
                                  , diedPlaceId))
            {
                Console.OUT.println(  "[Worker " + here + "]: I was asked to "
                                    + "restore " + Place(diedPlaceId)
                                    + " but cannot. " + "Crashing program.");

                Worker.killAll();
                Runtime.probe(); // To receive the killAll
                return;
            }

            AssertPrinter.println(  "[Worker " + here + "]: Queue-size "
                                  + "before restore = " + this.queue.size()
                                  + ", result before restore = "
                                  + this.queue.getResult().getResult());

            AssertPrinter.println(  "[Worker " + here + "]: backup = " + this.receiver.backup);

            AssertPrinter.println(  "[Worker " + here + "]: my queue: "
                                  + this.queue +  ".");

            this.queue.merge(this.receiver.backup);
            this.queue.mergeResult(this.receiver.backup);

            AssertPrinter.println(  "[Worker " + here + "]: Queue-size "
                                  + "after restore = " + this.queue.size()
                                  + ", result after restore = "
                                  + this.queue.getResult().getResult());

            this.receiver.backup = (null as Queue);
            this.sender.log = oldForthId;
            if (Long.MAX_VALUE < restore.tan)
            {
                this.receiver.tan = restore.tan;
                this.receiver.thieves = restore.thieves;
            }
            if (doNotWait)
            {
                this.mergeEmergencyMerges();
            }
        }
    }

    /**
     * We are asked whether we received loot from a specific place with a 
     * specific tan. If we did, we answer with gotOk; we answer with gotFalse
     * otherwise.
     * Role: thief
     * Sink: no, we have to send an answer.
     */
    @Inline public def gotCheck(val tan:Long
                                , val placeId:Long
                                , val srcId:Long) : void
    {
        val gotTasks:Boolean = (this.gotTasksFrom.keySet().contains(placeId)
                                && this.gotTasksFrom(placeId) >= tan);

        val st:PlaceLocalHandle[Worker[Queue, R]] = this.st;
        val me:Long = this.me;
        val src:Place = Place(srcId);
        if (gotTasks)
        {
            at (src) async
            {
                st().gotOk(tan, me);
            }
        }
        else // if (!gotTasks)
        {
            at (src) async
            {
                st().gotFalse(tan, me);
            }
        }
    }

    /**
     * Used in restore.
     * Role: back(diedPlace)
     * Sink: No
     */
    @Inline public def gotOk(val tan:Long, val srcId:Long) : void
    {
        this.timeout.remove(srcId, Timeout.GOT_ACK);
        this.openFEndManager.remove(tan, srcId);

        if (   !this.timeout.containsReason(Timeout.GOT_ACK)
            && !this.timeout.containsReason(Timeout.LNK_RES))
        {
            this.mergeEmergencyMerges();
        }
    }

    /**
     * One place did not received his share of loot. We cannot restore a valid
     * program state. We crash the program.
     * Role: back(diedPlace)
     * Sink: no, we have to kill every place via killAll.
     */
    @Inline public def gotFalse(val tan:Long, val placeId:Long) : void
    {
        // we crash the program anyway
        // this.timeout.remove(srcId, Timeout.GOT_ACK);
        val place:Place = Place(placeId);

        Console.OUT.println(  here + ": " + place
                            + " send an gotFalse for tan " + tan
                            + ". Crashing program.");

        Worker.killAll();
        Runtime.probe(); // To receive the killAll
    }

    /**
     * Kill all places
     * Role: any
     * Sink: no, we have to kill every place.
     */
    @Inline public static def killAll() : void
    {
        at (Place(0)) async {
            for (val place:Place in Place.places())
            {
                if (0 == place.id())
                {
                    continue;
                }

                at (place) async
                {
                    while (true)
                    {
                        Runtime.probe();
                    }
                }
            }

            for (val place:Place in Place.places())
            {
                if (0 == place.id())
                {
                    continue;
                }

                try
                {
                    at (place) @Uncounted async
                    {
                        Console.OUT.println(  "[Worker " + here
                                            + "]: Received KILL signal. Terminating.");

                        System.killHere();
                    }
                }
                catch (val c:CheckedThrowable)
                {}
            }
            while (1 < Place.places().filterDeadPlaces().size) {}
            
            Console.OUT.println("[Worker " + here + "]: I am the last survivor. Terminating now.");
            System.killHere();
        }
    }

    /**
     * We are requested to send the taskBag of a specified place to a specified
     * tan. If no TaskBag is found, we give a null-TaskBag.
     * Role: victim
     * Sink: no, we have to send the taskBag
     */
    @Inline public def linkResolve(val placeId:Long
                                   , val srcId:Long
                                   , val tans:List[Long]) : void
    {
        val taskGroups:List[Pair[Long, TaskBag]]
            = new ArrayList[Pair[Long, TaskBag]]();

        var size:Long = 0;
        for (val tan:Long in tans)
        {
            val entry:Pair[Long, TaskBag] = this.openManager.get(tan, placeId);
            taskGroups.add(entry);
            if (null != entry.second)
            {
                size += entry.second.size();
            }
            this.delOpen(tan, placeId);
            this.bfAck(tan, placeId, false);
        }
        val src:Place = Place(srcId);
        val st:PlaceLocalHandle[Worker[Queue, R]] = this.st;
        val me:Long = this.me;
        AssertPrinter.println(  "[Worker " + here + "]: sending "
                              + size + " as resolve to " + src + ".");
        at (src) async
        {
            st().linkTasks(me, taskGroups);
        }
    }

    /**
     * Remoe an entry from the openManager.
     * Role: victim
     * Sink: yes
     */
    @Inline public def delOpen( val tan:Long, val placeId:Long) : void
    {
        var leDelOpen:LEdelOpen = new LEdelOpen(System.nanoTime());
        this.logManager.startLE(leDelOpen);
        leDelOpen.tan = tan;
        leDelOpen.placeId = placeId;
        this.openManager.remove(tan, placeId);
        this.logManager.endLE();
    }

    /**
     * Link tasks in our links.
     * Role: back(thief)
     * Sink: yes
     */
    @Inline public def linkTasks(val srcId:long
                                , val taskGroups:List[Pair[Long, TaskBag]])
        : void
    {
        AssertPrinter.println("[Worker " + here + "]: got link from "
                              + Place(srcId) + ".");

        this.timeout.remove(srcId, Timeout.LNK_RES);
        for (entry:Pair[Long, TaskBag] in taskGroups)
        {
            if (null != entry.second)
            {
                this.emergencyMerges.push(entry.second);
            }
            this.linkManager.remove(entry.first, srcId);
        }

        if (   !this.timeout.containsReason(Timeout.GOT_ACK)
            && !this.timeout.containsReason(Timeout.LNK_RES))
        {
            this.mergeEmergencyMerges();
        }
    }

    public def mergeEmergencyMerges() : void
    {
        AssertPrinter.println(  "[Worker " + here + "]: merging all emergency "
                              + " merges.");

        while (!this.emergencyMerges.isEmpty())
        {
            val bag:TaskBag = this.emergencyMerges.pop();
            this.queue.merge(bag);
            this.state.empty = false;
        }

        assert (this.emergencyMerges.isEmpty())
               : here + ": after emergency-merge, but still merges left?!";

        // this.prot(0);
        val back:Long = this.back();

        val timeout:Timeout = new Timeout(back
                                          , Timeout.TO_ACK
                                          , Worker.TIMEOUT_TO);

        if (!this.timeout.add(timeout))
        {
            this.isDead(back);
        }

        if (this.timeout.containsReason(Timeout.RST_ACK))
        {
            this.state.answerRstAckAfterIAAck = true;
        }
        else
        {
            this.state.sendingBackup = true;
            this.sender.sendFullBackupTO(back);
            if (!this.state.active)
            {
                this.state.active = true;
                this.state.stealFailed = false;
                this.logger.startLive();
                this.logManager.startLE(new LEworking(System.nanoTime()));
                this.processStack();
                this.state.active = false;
                this.logger.stopLive();
                this.logManager.endLE();
                this.logger.nodesCount = this.queue.count();
            }
            // else
            // {
            //     this.processStack();
            // }
        }
    }

    /**
     * Receive Ack to a regular backup.
     * Role: sender of backup
     * sink: yes
     */
    @Inline public def regAck(val srcId:Long) : void
    {
        var leRegAck:LEregAck = new LEregAck(System.nanoTime());
        this.logManager.startLE(leRegAck);
        leRegAck.srcId = srcId;
        this.timeout.remove(srcId, Timeout.REG_ACK);
        AssertPrinter.println("[Worker " + here + "]: received regAck.");
        this.state.sendingBackup = false;
        // if (this.logger.timeReference + 10000000000 - System.nanoTime() < 0
        //     && (here.id() % 2 == 1))
        // {
        //     System.killHere();
        // }
        this.logManager.endLE();
    }

    /**
     * Receive Ack to a steal-backup-request.
     * Role: victim
     * sink: no, we have to distribute the loot.
     */
    @Inline public def stlAck(val tan:Long, val srcId:Long) : void
    {
        var leStlAck:LEstlAck = new LEstlAck(System.nanoTime());
        this.logManager.startLE(leStlAck);
        leStlAck.tan = tan;
        leStlAck.srcId = srcId;
        this.state.sendingBackup = false;
        this.timeout.remove(srcId, Timeout.STL_ACK);
        this.giveLootToThieves(tan);
        this.logManager.endLE();
    }

    @Inline private def giveLootToThieves(val tan:Long) : void
    {
        var leGiveLootToThieves:LEgiveLootToThieves
            = new LEgiveLootToThieves(System.nanoTime());
        this.logManager.startLE(leGiveLootToThieves);
        leGiveLootToThieves.tan = tan;
        leGiveLootToThieves.tanDescriptorSize = this.tanDescriptor.size();
        val me:Long = this.me;
        var allAlive:Boolean = true;
        for (var index:Long = 0; index < this.tanDescriptor.size(); ++index)
        {
            val entry:Pair[Long, TaskBag] = this.tanDescriptor.get(index);
            var thiefId:Long = entry.first;
            leGiveLootToThieves.tanThiefIds = leGiveLootToThieves.tanThiefIds + thiefId + ":";
            val thief:Place = Place(thiefId);
            val loot:TaskBag = entry.second;
            leGiveLootToThieves.lootSize = leGiveLootToThieves.lootSize + loot.size() + ":";
            val st:PlaceLocalHandle[Worker[Queue, R]] = this.st;

            val timeout:Timeout = new Timeout(thiefId
                                              , Timeout.BF_ACK
                                              , Worker.TIMEOUT_BFACK);

            if (!this.timeout.add(timeout))
            {
                allAlive = false;
                this.tanDescriptor.removeAt(index--);
                this.queue.merge(entry.second);
            }
            /*
             * !! IMPORTANT!!
             * This is NOT uncounted. This can restart the worker.
             */
            at (thief) async
            {
                st().give(tan, me, loot);
            }
        }
        if (!allAlive)
        {
            if (this.state.canSendBackup())
            {
                val backupId:Long = this.backForth.getBack();

                val timeout:Timeout
                    = new Timeout(backupId, Timeout.TO_ACK, Worker.TIMEOUT_TO);

                this.state.sendingBackup = true;
                this.sender.sendFullBackupTO(backupId);
            }
            else
            {
                this.state.sendBackupTO = true;
            }
        }
        this.logManager.endLE();
    }

    @Inline public def toAck(val srcId:Long) : void
    {
        var leToAck:LEtoAck = new LEtoAck(System.nanoTime());
        this.logManager.startLE(leToAck);
        leToAck.srcId = srcId;
        this.timeout.remove(srcId, Timeout.TO_ACK);
        this.state.sendingBackup = false;
        val st:PlaceLocalHandle[Worker[Queue, R]] = this.st;
        val me:Long = this.me;
        AssertPrinter.println("[Worker " + here + "]: got toAck from " + Place(srcId) + ".");
        if (this.state.repeatTOBackup)
        {
            this.state.repeatTOBackup = false;
            this.mergeEmergencyMerges();
        }
        else // if (!this.state.repeatTOBackup)
        {
            assert (-1 < this.state.answerRestore) : "answerRestore is -1.";
            val dst:Place = Place(this.state.answerRestore);
            this.state.answerRestore = -1;
            AssertPrinter.println("[Worker " + here + "]: sending rstAck to " + dst + ".");
            at (dst) async
            {
                st().rstAck(me);
            }
        }
        this.logManager.endLE();
    }

    @Inline public def iaAck(val srcId:Long) : void
    {
        var leIaAck:LEiaAck = new LEiaAck(System.nanoTime());
        this.logManager.startLE(leIaAck);
        leIaAck.srcId = srcId;
        this.timeout.remove(srcId, Timeout.IA_ACK);
        this.state.emergency = false;
        this.state.restoreInProcess = -1;
        this.state.sendingBackup = false;

        val openFEnd:ArrayList[Pair[Long, Long]]
        = this.openFEndManager.openFEnd;

        val me:Long = this.me;

        for (var index:Long = 0; index < openFEnd.size(); ++index)
        {
            entry:Pair[Long, Long] = openFEnd.get(index);
            val tan:Long = entry.first;
            val victimId:Long = entry.second;
            val victim:Place = new Place(victimId);
            val st:PlaceLocalHandle[Worker[Queue, R]] = this.st;
            at (victim) @Uncounted async
            {
                st().bfAck(tan, me);
            }
            openFEnd.removeAt(index--);
        }

        val timeouts = this.timeout.waitingFor;
        val st:PlaceLocalHandle[Worker[Queue, R]] = this.st;
        for (var index:Long = 0; index < timeouts.size(); ++index)
        {
            val timeout:Timeout = timeouts.get(index);
            if (Timeout.F_END == timeout.reason)
            {
                val victimId:Long = timeout.placeId;
                val victim:Place = Place(victimId);
                val tan:Long = this.gotTasksFrom.get(victimId);
                at (victim) @Uncounted async
                {
                    st().bfAck(tan, me, false);
                }
                this.openFEndManager.remove(tan, victimId);
            }
        }

        if (0 < this.tanDescriptor.size())
        {
            this.giveLootToThieves(this.currentTan);
        }

        if (this.state.answerRstAckAfterIAAck)
        {
            this.state.answerRstAckAfterIAAck = false;
            val src:Place = Place(this.state.answerRestore);
            this.state.answerRestore = -1;
            at (src) async
            {
                st().rstAck(me);
            }
        }
        if (this.state.blockTOBackup)
        {
            this.state.emergency = true;
            // 1st restore as back has not yet sent a toBackup and 2nd restore
            //as forth was able to complete -> continue normal
            if (   this.timeout.containsReason(Timeout.GOT_ACK)
                || this.timeout.containsReason(Timeout.LNK_RES))
            {
                this.state.blockTOBackup = false;
                this.logManager.endLE();
                return;
            }
            // 1st restore as back has send toBackup and 2nd restore
            // as forth has finished -> repeat toBackup as back
            else if (this.timeout.containsReason(Timeout.TO_ACK))
            {
                this.state.blockTOBackup = false;
                this.state.repeatTOBackup = true;
            }
            // 1st restore as back has been finished and 2nd restore as forth
            // has been finished -> repeat toBackup
            else // if (   !this.timeout.containsReason(Timeout.GOT_ACK)
                 //     && !this.timeout.containsReason(Timeout.LNK_RES)
                 //     && !this.timeout.containsReason(Timeout.TO_ACK))
            {
                this.state.repeatTOBackup = false;
                this.state.blockTOBackup = false;
                this.mergeEmergencyMerges();
            }
        }
        if (this.state.repeatStlReq)
        {
            this.stlAck(this.currentTan, srcId);
        }
        this.logManager.endLE();
    }

    @Inline public def rstAck(val srcId:Long) : void
    {
        var leRstAck:LErstAck = new LErstAck(System.nanoTime());
        this.logManager.startLE(leRstAck);
        leRstAck.srcId = srcId;
        this.timeout.remove(srcId, Timeout.RST_ACK);
        this.backForth.updateBack();

        // The new backup place was able to send the rstAck, but died
        // afterwards. In this case, we call isDead with its ID and do NOT
        // proceed to write the IA-backup.
        if (srcId != this.backForth.getBack())
        {
            this.isDead(srcId);
            this.logManager.endLE();
            return;
        }
        AssertPrinter.println("[Worker " + here + "]: got rstAck from " + Place(srcId) + ", sending iaReq.");
        this.iaReq();
        this.logManager.endLE();
    }

    public def isDead(val deadPlaceId:Long)
    {
        val backupId:Long = this.backForth.getBack();

        if ((deadPlaceId == this.state.restoreInProcess))
        {
            return;
        }

        // I am already fulfilling a role as back(diedPlace) and now
        // additionally as forth(anotherDiedPlace). Delay the TO-backup to after
        // IAAck. If not possible, repeat TO-Backup.
        if (   this.timeout.containsReason(Timeout.GOT_ACK)
            || this.timeout.containsReason(Timeout.LNK_RES)
            || this.timeout.containsReason(Timeout.TO_ACK))
        {
            this.state.blockTOBackup = true;
        }

        assert (Worker.checkLog(this.me, backupId, deadPlaceId))
               :  here + ": something went horrible, horrible wrong. "
                + "I was notified about " + Place(deadPlaceId) + "'s death, but "
                + "i am not responsible.";

        val idx:Long = this.timeout
                           .waitingFor
                           .indexOf(new Timeout(deadPlaceId
                                                , Timeout.RST_ACK));

        // We have send out a restore request to the died place.
        // We cannot expect to receive a RST_ACK.
        // We have to send a restore-request to back(diedPlace) (which hopefully
        // will be our next backup-place)
        
        if (-1 < idx)
        {
            this.timeout.waitingFor.removeAt(idx);
        }

        // this.backForth.resetForth();

        // Either our backup-place died or the place, we tried to get as our
        // new-backup-place died.
        if ((deadPlaceId == backupId) || (-1 < idx))
        {
            this.state.emergency = true;
            this.state.restoreInProcess = deadPlaceId;
            var newBackupPlaceFound:Boolean = false;
            while (!newBackupPlaceFound)
            {
                val newBackup:Place = Place.places().next(here);
                val newBackupId:Long = newBackup.id();

                val timeout:Timeout = new Timeout(newBackupId
                                                  , Timeout.RST_ACK
                                                  , Worker.TIMEOUT_RST);

                if (!this.timeout.add(timeout))
                {
                    continue;
                }

                if (this.state.repeatStlReq)
                {
                    val timeout2:Timeout = new Timeout(newBackupId
                                                       , Timeout.STL_ACK
                                                       , Worker.TIMEOUT_STEAL);

                    if (!this.timeout.add(timeout2))
                    {
                        this.timeout.remove(timeout.placeId, timeout.reason);
                        continue;
                    }
                }
                newBackupPlaceFound = true;
                val st:PlaceLocalHandle[Worker[Queue, R]] = this.st;
                val me = this.me;

                var myThieves:List[Long] = new ArrayList[Long]();
                for (entry:Pair[Long, TaskBag] in this.tanDescriptor)
                {
                    myThieves.add(entry.first);
                }

                val thieves:List[Long] = (  myThieves.isEmpty() 
                                          ? null : myThieves);

                val tan:Long
                    = (myThieves.isEmpty() ? Long.MAX_VALUE : this.currentTan);

                at (newBackup) async
                {
                    st().restore(tan, deadPlaceId, me, thieves);
                }
            }
        }
    }

    /**
     * Receive an ack from a back(thief) to a give.
     * Role: victim
     * sink: no, we have to send an fEdn to the thief and eventually a bvEnd to
     * our backup-place.
     */
    @Inline public def bfAck(val tan:Long, val thiefId:Long) : void
    {
        this.bfAck(tan, thiefId, true);
    }

    @Inline public def bfAck(val tan:Long
                             , val thiefId:Long
                             , val contactThief:Boolean) : void
    {
        var leBfAck:LEbfAck = new LEbfAck(System.nanoTime());
        this.logManager.startLE(leBfAck);
        leBfAck.tan = tan;
        leBfAck.thiefId = thiefId;
        leBfAck.contactThief = contactThief;
        val me:Long = this.me;
        if (contactThief)
        {
            val thief:Place = Place(thiefId);
            val st:PlaceLocalHandle[Worker[Queue, R]] = this.st;
            //leBfAck.time(0) = System.nanoTime();
            at (thief) @Uncounted async
            {
                st().fEnd(tan, me);
            }
        }
        //leBfAck.time(1) = System.nanoTime();
        this.timeout.remove(thiefId, Timeout.BF_ACK);
        //leBfAck.time(2) = System.nanoTime();
        for (var index:Long = 0; index < this.tanDescriptor.size(); ++index)
        {
            val entry:Pair[Long, TaskBag] = this.tanDescriptor(index);
            if (thiefId == entry.first)
            {
                this.tanDescriptor.removeAt(index);
                break;
            }
        }
        //leBfAck.time(3) = System.nanoTime();
        
        AssertPrinter.println(  "[Worker " + here + "]: Thief " + Place(thiefId)
                              + " has bfAck'ed");

        //leBfAck.time(4) = System.nanoTime();
        
        AssertPrinter.println(  "[Worker " + here + "]: I have " 
                              + this.tanDescriptor.size()
                              + " thieves still open.");

        //leBfAck.time(5) = System.nanoTime();
        if (tanDescriptor.isEmpty())
        {
            //leBfAck.time(6) = System.nanoTime();
            AssertPrinter.println(  "[Worker " + here 
                                  + "]: sending bvEnd, no more thieves open.");

            //leBfAck.time(7) = System.nanoTime();
            val st:PlaceLocalHandle[Worker[Queue, R]] = this.st; //Bugfix
            val backup:Place = Place(this.back());
            //leBfAck.time(8) = System.nanoTime();
            this.sleBfAck.setST();
            at (backup) @Uncounted async
            {
                st().bvEnd(me);
            }
            this.sleBfAck.setET();
            //leBfAck.time(9) = System.nanoTime();
            // this.state.setV(2);
        }
        //leBfAck.time(10) = System.nanoTime();
        this.logManager.endLE();
    }

    /**
     * Receive end-message for steal protocol.
     * Role: thief
     * Sink: yes.
     */
    @Inline public def fEnd(val tan:Long, val srcId:Long) : void
    {
        var leFEnd:LEfEnd = new LEfEnd(System.nanoTime());
        this.logManager.startLE(leFEnd);
        leFEnd.tan = tan;
        leFEnd.srcId = srcId;
        this.openFEndManager.remove(tan, srcId);
        this.logManager.endLE();
    }

    /**
     * Receive end-message for steal protocol
     * Role: back(victim)
     * Sink: yes
     */
    @Inline public def bvEnd(val srcId:Long) : void
    {
        var leBvEnd:LEbvEnd = new LEbvEnd(System.nanoTime());
        this.logManager.startLE(leBvEnd);
        leBvEnd.srcId = srcId;
        if (this.forth() == srcId)
        {
            this.receiver.bvEnd();
            this.state.valid = true;
        }
        this.logManager.endLE();
    }

    /**
     * start (cyclic) monitor-process.
     * Role: any
     * Sink: no, we have to check at least one remote place.
     */
    @Inline public def monitor() : void
    {
        // AssertPrinter.println("[Worker " + here + "]: starting monitor.");
        this.monitor.monitor();
    }

    /**
     * Get the id of our backup-place.
     * Role: any
     * Sink: yes
     */
    @Inline public def back() : Long
    {
        return (this.backForth.getBack());
    }

    /**
     * Get the id of our backedup-place.
     * Role: any
     * Sink: yes
     */
    @Inline public def forth() : Long
    {
        return (this.backForth.getForth());
    }

//     /**
//      * Protocol current queue-state, if it is a new min-state.
//      * Role: any
//      * Sink: yes
//      */
//     @Inline public def prot(val cycle:Long) : void
//     {
//         if (   (this.queue.size() <= this.sender.minSize() + 1)
//             || ((0 == cycle) && (0n == this.currentK))
//             || this.state.emergency)
//         {
//             this.sender.prot();
//         }
// 
//         // if ((0 == cycle) && (0n == this.currentK))
//         // {
//         //     AssertPrinter.println("[Worker " + here + "]: first cycle prot.");
//         // }
//         // AssertPrinter.println("[Worker " + here + "]: prot.");
//     }

    /**
     * Process merges, which could not be processed at their time of arrival.
     * This is done by re-calling this.give (as the victim would) with the
     * stored parameters.
     * Role: thief
     * Sink: no, we have to send link(s) to our backup-place.
     */
    @Inline public def processMerges() : void
    {
        if (0 == this.message.recordedMerges.size())
        {
            return;
        }
        //Logevent processMerges start. Gehört zum Stealing
        var leProcessMerges:LEprocessMerges = new LEprocessMerges(System.nanoTime());
        this.logManager.startLE(leProcessMerges);
        
        val undone:SimpleStack[Pair[Long, Pair[Long, TaskBag]]]
            = new SimpleStack[Pair[Long, Pair[Long, TaskBag]]]();

        while (this.message.recordedMerges.size() > 0)
        {
            val entry:Pair[Long, Pair[Long, TaskBag]]
                = this.message.recordedMerges.pop();

            val victimId:Long = entry.first;
            val tan:Long = entry.second.first;
            val loot:TaskBag = entry.second.second;
            if (!this.give(tan, victimId, loot, true))
            {
                undone.push(entry);
            }
        }

        if (!undone.isEmpty())
        {
            this.message.recordedMerges.pushAll(undone);
        }
        // else // if(undone.isEmpty())
        // {
        //     AssertPrinter.println(  "[Worker " + here
        //                           + "]: no more merges recorded.");
        // }
        //Logevent processMerges ende.
        this.logManager.endLE();
    }

    @Inline public def resetK()
    {
        this.currentK = 0n;
    }

    @Inline public def resetR()
    {
        this.currentR = 0n;
    }

    public def ghostActivity() : void
    {
        this.processRecorded();
        this.state.ghost = false;
    }

    @Inline public static def checkLog(val lower:Long
                                       , val upper:Long
                                       , val id:Long) : Boolean
    {
        if (lower <= upper)
        {
            if ((id >= lower) && (id <= upper))
            {
                return (true);
            }
        }
        else // if (lower > upper)
        {
            if ((id >= lower) || (id <= upper))
            {
                return (true);
            }
        }
        return (false);
    }

    /**
     * 
     */
    @Inline public def checkJoins()
    {
        if (   0 == here.id()
            && Place.numPlaces() > this.P
            && !this.timeout.containsReason(Timeout.PROCESS_JOINS))
        {
            this.timeout.add(new Timeout(0, Timeout.PROCESS_JOINS, TIMEOUT_PROCESS_JOINS));
        }
    }

    /**
     * Method to check whether new places have joined the computation.
     * If new places have joined, they will be added to {@link #st}.
     */
    @Inline public def processJoins()
    {
        // new places arrived
        // if (Place.numPlaces() > this.P)
        // {
            val oldP:Long = this.P;
            val newPlaces:ArrayList[Place] = new ArrayList[Place]();
            val makeEmptyQueue:()=>Queue = ()=>{ return (this.queue.makeEmptyQueue() as Queue); };

            val makeWorker:()=>Worker[Queue, R] = () => 
            {
                return new Worker[Queue, R](  makeEmptyQueue
                                            , this.n
                                            , this.w
                                            , this.l
                                            , this.m
                                            , this.k
                                            , this.r);
            };

            val test:()=>Worker[Queue, R] = ()=>
            {
                Console.OUT.println(here + ": called");
                worker:Worker[Queue, R] = makeWorker();
                return (worker);
            };

            finish
            {
                for (var id:Long = this.P; id < Place.numPlaces(); ++id)
                {
                    val newPlace:Place = Place(id);
                    newPlaces.add(newPlace);
                    Console.OUT.println("[Worker " + here + "]: adding " + newPlace + ".");
                    PlaceLocalHandle.addPlace(this.st, Place(id), test);

                    at (newPlace) async
                    {
                        st().init(st);
                    }

                    Console.OUT.println("[Worker " + here + "]: " + id + " added.");
                }
            }

            val newLifelineList:HashMap[Long, ArrayList[Long]] = calculateAllLifelines(  this.P
                                                                                       , this.l);

            val st:PlaceLocalHandle[Worker[Queue, R]] = this.st;
            for (val p:Place in Place.places())
            {
                if (!p.isDead())
                {
                    val isNew:Boolean = newPlaces.contains(p);
                    val oldLifelines = this.lifelineList.get(p.id());
                    val newLifelines = newLifelineList.get(p.id());
                    if (    (oldLifelines == newLifelines)
                        ||  (   (null != oldLifelines)
                             && (null != newLifelines)
                             && oldLifelines.containsAll(newLifelines)
                             && newLifelines.containsAll(oldLifelines)))
                    {
                        at (p) async
                        {
                            st().P = Place.numPlaces();
                            if (isNew)
                            {
                                st().main();
                            }
                        }
                    }
                    else
                    {
                        at (p) async
                        {
                            st().P = Place.numPlaces();
                            if (isNew)
                            {
                                st().main();
                            }
                            updateLifelines(newLifelines);
                        }
                    }
                }
            }

        // }
    }

    @Inline public def updateLifelines(val lifelines:ArrayList[Long]) : void
    {
        this.lifelines = lifelines;
        if (nextVictim > 0)
        {
            nextVictim = 0;
        }
    }
}
