package de.uks.plm.ftglb.actor.managers;

import x10.compiler.Inline;
import x10.compiler.NonEscaping;

import x10.util.ArrayList;
import x10.util.List;
import x10.util.Pair;

import de.uks.plm.ftglb.actor.Worker;
import de.uks.plm.ftglb.actor.TaskBag;
import de.uks.plm.ftglb.actor.TaskQueue;
import de.uks.plm.ftglb.actor.dataStructures.Restore;

import de.uks.plm.ftglb.actor.dataStructures.BackupData;

import de.uks.plm.utils.SimpleStack;

/**
 * Manage a stack for each method-type.
 */
public class MessageManager[Queue, R] {Queue <: TaskQueue[Queue, R]}
{
    public var worker:Worker[Queue, R];
    public val thieves:SimpleStack[Long];
    public val lifelineThieves:SimpleStack[Long];
    public val lifelinesActivated:List[Long];

    /**
     * 1st Pair:
     *     1st Entry: source id
     *     2nd Entry: 
     *         1st Entry: tan
     *         2nd entry: loot
     */
    public val recordedMerges:SimpleStack[Pair[Long, Pair[Long, TaskBag]]];

    public val recordedRestores:SimpleStack[Restore];

    public var haveBackupRecorded:Boolean;
    public var recordedBackup:Pair[Long, BackupData[Queue, R]];

    /**
     * Queue for restore-requests.
     * 1st element of the pair is the dead place
     * 2nd elemente of the pair is the source, which send the request
     */
    public val restore:SimpleStack[Pair[Long, Long]];

    public def this() : MessageManager[Queue, R]
    {
        this.thieves = new SimpleStack[Long]();
        this.lifelineThieves = new SimpleStack[Long]();
        this.restore = new SimpleStack[Pair[Long, Long]]();
        this.lifelinesActivated = new ArrayList[Long](Place.numPlaces());

        this.recordedMerges
            = new SimpleStack[Pair[Long, Pair[Long, TaskBag]]]();

        this.recordedRestores = new SimpleStack[Restore](Place.numPlaces());
        this.haveBackupRecorded = false;
        this.recordedBackup = new Pair[Long, BackupData[Queue, R]](-1, null);
    }

    /**
     * Record a backup to process later.
     * role: back
     * sink: yes
     */
    public def recordBackup(val srcId:Long, val data:BackupData[Queue, R])
        : void
    {
        assert (!this.haveBackupRecorded)
               : here + ": holding already a recorded backup!";

        this.haveBackupRecorded = true;
        this.recordedBackup = new Pair[Long, BackupData[Queue, R]](srcId, data);
        this.checkGhostMode();
    }

    public def recordRestore (val tan:Long
                              , val diedPlace:Long
                              , val srcId:Long
                              , val thieves:List[Long]) : void
    {
        val entry:Restore = new Restore(tan, diedPlace, srcId, thieves);
        this.recordedRestores.push(entry);
        this.checkGhostMode();
    }

    /**
     * Get the recorde backup(if any).
     * role: back
     * sink: yes
     */
    public def getBackup() : Pair[Long, BackupData[Queue, R]]
    {
        if (this.haveBackupRecorded)
        {
            this.haveBackupRecorded = false;
            return (this.recordedBackup);
        }
        /*
         * Since it is impossible to access template parameters for static
         * attributes, we have to create the "no backup to return" attribute
         * every time.
         */
        return (new Pair[Long, BackupData[Queue, R]](-1, null));
    }

    public def getRestore() : Restore
    {
        assert (!this.recordedRestores.isEmpty())
               : here + ": no restore recorded.";

        return (this.recordedRestores.pop());
    }

    /**
     * Record a merge to merge later.
     * role: thief
     * sink: yes
     */
    @Inline public def recordMerge(val tan:Long
                                   , val srcId:Long
                                   , val loot:TaskBag) : void
    {
        val innerEntry:Pair[Long, TaskBag]
            = new Pair[Long, TaskBag](tan, loot);

        val entry:Pair[Long, Pair[Long, TaskBag]]
            = new Pair[Long, Pair[Long, TaskBag]](srcId, innerEntry);

        this.recordMerge(entry);
        this.checkGhostMode();
    }

    /**
     * Record a merge to merge later.
     * role: thief
     * sink: yes
     */
    @Inline public def recordMerge(val entry:Pair[Long, Pair[Long, TaskBag]])
        : void
    {
        this.recordedMerges.push(entry);
        this.checkGhostMode();
    }

    /**
     * Get the recorded merge.
     * Throws an exception, if none is recorded.
     * role: thief
     * sink: yes
     */
    @Inline public def getMerge() : Pair[Long, Pair[Long, TaskBag]]
    {
        return (this.recordedMerges.pop());
    }

    /**
     * Activate a lifeline
     * role: thief
     * sink: yes
     */
    @Inline public def activateLifeline(val id:Long) : void
    {
        this.lifelinesActivated.add(id);
    }

    /**
     * Query whether specified lifeline is activated.
     * role: thief
     * sink: yes
     */
    @Inline public def isLifelineActive(val id:Long) : Boolean
    {
        return (this.lifelinesActivated.contains(id));
    }

    /**
     * Deactivate a lifeline.
     * role: thief
     * sink: yes
     */
    @Inline public def deactivateLifeline(val id:Long) : void
    {
        assert (this.lifelinesActivated.contains(id))
               : here + ": cannot deactivate a lifeline, that is not active!";

        this.lifelinesActivated.remove(id);
    }

    @Inline public def checkGhostMode()
    {
        if (!(this.worker.state.active || this.worker.state.ghost))
        {
            this.worker.state.ghost = true;
            async
            {
                this.ghostMode();
            }
        }
    }

    @Inline public def ghostMode()
    {
        this.worker.processRecorded();
        this.worker.state.ghost = false;
    }
}
