package de.uks.plm.ftglb.actor.managers;

import x10.compiler.Inline;

import x10.util.ArrayList;
import x10.util.Collection;
import x10.util.List;
import x10.util.Pair;

import de.uks.plm.ftglb.actor.TaskBag;
import de.uks.plm.ftglb.actor.TaskQueue;
import de.uks.plm.ftglb.actor.Worker;

import de.uks.plm.utils.AssertPrinter;

/**
 * This class manages adding and removing Transaction-numbers. It is used on
 * the side of the victim.
 *
 * @author Marco Bungart
 */
public class OpenManager[Queue, R] {Queue <: TaskQueue[Queue, R]}
{
    public static val fillOpen:(val idx:Long)=>List[Pair[Long, TaskBag]]
        = (val idx:Long) => 
        {
            return new ArrayList[Pair[Long, TaskBag]]();
        };

    /**
     * Default-value returned, if a specified pair is not found with
     * @link{#get(Long, Long) get(Long place, Long tan)}.
     */
    public static val NOT_FOUND_PAIR:Pair[Long, TaskBag]
        = new Pair[Long, TaskBag](Long.MIN_VALUE, null);

    /**
     * Structure of open:
     * 
     * Array-index = Place-id
     * 1st entry of Pair: tan
     * 2nd entry of Pair: loot
     */
    private var open:Rail[List[Pair[Long, TaskBag]]];

    /**
     * Counter for tans. The current value is the next usable tan.
     * If this value overflows, an assertion will fail.
     */
    private var nextTan:Long = Long.MIN_VALUE;

    /**
     * Constructor.
     */
    public def this() : OpenManager[Queue, R]
    {
        val P:Long = Place.numPlaces();
        this.open = new Rail[List[Pair[Long, TaskBag]]](P);

        for (p:Long in 0..(P-1))
        {
            this.open(p) = new ArrayList[Pair[Long, TaskBag]]();
        }
    }

    /**
     * Checks, whether this tan is still open. All transactios from this place
     * to other places are checked.
     * 
     * @param tan the tan so search for
     *
     * @return true, if the tan is found, false if not.
     */
    @Inline public def containsTan(val tan:Long) : Boolean
    {
        for (val list:List[Pair[Long, TaskBag]] in this.open)
        {
            for (val entry:Pair[Long, TaskBag] in list)
            {
                if (tan == entry.first)
                {
                    return (true);
                }
            }
        }
        return (false);
    }

    /**
     * This method is internally used to add tans with their corresponding
     * @link{TaskBag TaskBags} to a transaction from this place to another
     * place.
     *
     * @param place the id of the receiving place.
     * @param tan the tan.
     * @param loot the corresponding loot.
     */
    @Inline private def add(val tan:Long
                            , val place:Long
                            , val loot:TaskBag) : void
    {
        this.add(place, new Pair[Long, TaskBag](tan, loot));
    }

    /**
     * This method is internally used to add a Pair[Long, TaskBag] as a
     * transaction from this place to another place.
     *
     * @param place the id of the receiving place.
     * @param tanEntry the Pair to add. First component is the tan,
     *        second component is the @link{TaskBag TaskBag}.
     */
    @Inline private def add(val place:Long
                           , val tanEntry:Pair[Long, TaskBag]) : void
    {
        while (this.open.size - 1 < place)
        {
            grow();
        }
        this.open(place).add(tanEntry);
    }

    @Inline private def grow() : void
    {
        if (Long.MAX_VALUE == this.open.size)
        {
            return;
        }

        val oldSize = this.open.size;
        val newSize:Long = oldSize * 2 < 0 ? Long.MAX_VALUE : oldSize * 2;

        val newOpen:Rail[List[Pair[Long, TaskBag]]]
            = new Rail[List[Pair[Long, TaskBag]]](newSize, fillOpen);

        Rail.copy(this.open, 0, newOpen, 0, this.open.size);
        this.open = newOpen;

    }
    /**
     * A method to add multiple loots to multiple transactions. The tan is
     * generated in the process. If @link{#nextTan nextTan} overflows, an
     * assertion will fail.
     *
     * @param addList this list is a List[Pair[Long, TaskBag]]. The first
     *                component of the Pair specifies the id of the place,
     *                the second component specifies the loot.
     *
     * @return the tan for this transactions.
     */
    @Inline public def add(val addList:List[Pair[Long, TaskBag]]) : Long
    {
        val tan:Long = this.getNextTan();

        val thieves:List[Long] = new ArrayList[Long]();
        for (toAdd:Pair[Long, TaskBag] in addList)
        {
            val place:Long = toAdd.first;
            val loot:TaskBag = toAdd.second;

            this.add(tan, place, loot);
        }
        return (tan);
    }

    /**
     * Get all open tans of the specified place. Open tans are those ones, which
     * are currently in transaction or not yet resolved.
     * 
     * 
     * @param place the place, from which one wished to obtain the open tans
     * @return a List[Long] of the open tans.
     */
    @Inline public def getOpenTans(val place:Long) : List[Long]
    {
        val tans:ArrayList[Long] = new ArrayList[Long](this.open(place).size());

        for (tuple:Pair[Long, TaskBag] in this.open(place))
        {
            tans.add(tuple.first);
        }

        return (tans);
    }

    /**
     * Get all open tans, which are less or euqal to upperTan from the specified
     * place.
     * Remark: specifying upperTan as Long.MAX_VALUE will get the same result
     * as calling @link{#getOpenTans(Long)}.
     * 
     * @param place the place, from which one wished to obtain the open tans
     * @param upperTan the limit to determin, which tans to get.
     * 
     * @return a List[Long] of the open tans less or equal to upperTan.
     */
    @Inline public def getOpenTansUpTo(val upperTan:Long
                                       , val place:Long) : List[Long]
    {
        val list:List[Long] = this.getOpenTans(place);

        val selector:(Long) => Boolean = (val tan:Long) => 
                                         {
                                             return (tan > upperTan);
                                         };

        list.removeAllWhere(selector);
        return (list);
   }

    /**
     * Gets a specified entry from @link{#open}.
     *
     * @param place the id of the place
     * @param tan the tan to look for
     *
     * @return if it is found, the entry with the corresponding tan,
     *         otherwise it returns a @link{java.util.Pair Pair} with the 
     *         @link{java.util.Pair#first first} Component set to
     *         @link{x10.lang.Long#MIN_VALUE} and the 
     *         @link{java.util.Pair#second second} component set to null.
     *         This is neccessary since @link{java.util.Pair} is a struct and
     *         therefore cannot be set to null.
     */
    public def get(val tan:Long, val place:Long) : Pair[Long, TaskBag]
    {
        var value:Pair[Long, TaskBag] = new Pair[Long, TaskBag](tan, null);

        for (entry:Pair[Long, TaskBag] in this.open(place))
        {
            if (tan == entry.first)
            {
                value = entry;
            }
        }
        return (value);
    }

    /**
     * Method to generate next tan.
     * If the tan overflows, an assertion will fail.
     */
    public def getNextTan() : Long
    {
        assert (Long.MAX_VALUE != this.nextTan) : here + ": nextTan overflew!";

        return (this.nextTan++);
    }

    /**
     * This method removes a tan from the list of transaction of the specified
     * place.
     *
     * @param place the id of the receiving place
     * @param tan the tan to delete
     *
     * @return true, if the tan is found, false if not.
     */
    @Inline public def remove(val tan:Long, val place:Long) : Boolean
    {
        AssertPrinter.println("[OpenManager " + here + "] removing (" + tan
                              + ", " + place + ") from open.");
        for (var index:Long = 0; index < this.open(place).size(); ++index)
        {
            val entry:Pair[Long, TaskBag] = this.open(place)(index);
            if (tan == entry.first)
            {
                this.open(place).removeAt(index);
                AssertPrinter.println(  "[OpenManager " + here
                                      + "]: found entry.");
                return (true);
            }
        }
        return (false);
    }

    /**
     * Remove a set of tans of a place from the openManager.
     * sink: yes
     */
    @Inline public def remove(val place:Long, val tans:ArrayList[Long]) : void
    {
        for (var oIndex:Long = 0; oIndex < this.open(place).size(); ++oIndex)
        {
            val oEntry:Pair[Long, TaskBag] = this.open(place)(oIndex);
            for (var tIndex:Long = 0; tIndex < tans.size(); ++tIndex)
            {
                if (tans(tIndex) == oEntry.first)
                {
                    this.open(place).removeAt(oIndex);
                    tans.removeAt(tIndex);
                    oIndex--;
                    tIndex--;
                    if (tans.size() == 0)
                    {
                        return;
                    }
                    continue;
                }
            }
        }

        assert (false) : here + ": not all tans in open found!";
    }
}
